import logging
from time import sleep
from unittest import TestCase, skipIf

from parameterized import parameterized, param

from hvac import exceptions
from tests import utils
from tests.utils.hvac_integration_test_case import HvacIntegrationTestCase


@skipIf(
    utils.vault_version_lt("0.10.0"),
    "KV version 2 secret engine not available before Vault version 0.10.0",
)
class TestKvV2(HvacIntegrationTestCase, TestCase):
    DEFAULT_MOUNT_POINT = "kvv2"

    def setUp(self):
        super().setUp()
        self.client.sys.enable_secrets_engine(
            backend_type="kv",
            path=self.DEFAULT_MOUNT_POINT,
            options=dict(version=2),
        )

        # We occasionally see issues with the newly enabled secrets engine not becoming available in time for our test cases.
        # So we wait for it to show up in the mounted secrets engines list here before proceeding.
        path = f"{self.DEFAULT_MOUNT_POINT}/"
        attempts = 0
        while (
            attempts < 25
            and path not in self.client.sys.list_mounted_secrets_engines()["data"]
        ):
            attempts += 1
            logging.debug(
                "Waiting 1 second for KV V2 secrets engine under path {path} to become available...".format(
                    path=self.DEFAULT_MOUNT_POINT,
                )
            )
            sleep(1)

    def tearDown(self):
        self.client.sys.disable_secrets_engine(path=self.DEFAULT_MOUNT_POINT)
        super().tearDown()

    @parameterized.expand(
        [
            ("no parameters",),
            ("set max versions", 1),
            ("set cas required", 10, True),
            ("set max versions and cas required", 17, True),
            ("set delete_version_after to thirty minutes", 10, None, "30m0s"),
        ]
    )
    def test_configure_and_read_configuration(
        self, test_label, max_versions=10, cas_required=None, delete_version_after="0s"
    ):
        configure_arguments = dict(
            delete_version_after=delete_version_after,
            mount_point=self.DEFAULT_MOUNT_POINT,
        )
        if max_versions is not None:
            configure_arguments["max_versions"] = max_versions
        if cas_required is not None:
            configure_arguments["cas_required"] = cas_required
        self.client.secrets.kv.v2.configure(**configure_arguments)
        read_configuration_response = self.client.secrets.kv.v2.read_configuration(
            mount_point=self.DEFAULT_MOUNT_POINT,
        )
        logging.debug("read_configuration_response: %s" % read_configuration_response)
        self.assertEqual(
            first=max_versions,
            second=read_configuration_response["data"]["max_versions"],
        )
        self.assertEqual(
            first=cas_required or False,
            second=read_configuration_response["data"]["cas_required"],
        )
        if utils.vault_version_ge("1.5.0"):
            # delete_version_after wasn't returned in this response when unset before 1.5.0
            self.assertEqual(
                first=delete_version_after,
                second=read_configuration_response["data"]["delete_version_after"],
            )

    @parameterized.expand(
        [
            (
                "nonexistent secret",
                "no-secret-here",
                None,
                False,
                exceptions.InvalidPath,
            ),
            ("read secret version 2 back", "top-secret", 2, 2),
            ("read secret version 1 back", "top-secret", 1, 5),
            ("read current secret version", "top-secret", None, 10),
            ("read current secret version", "top-secret", None, 15),
        ]
    )
    def test_read_secret_version(
        self, test_label, path, version=None, write_secret_before_test=True, raises=None
    ):
        if write_secret_before_test:
            for num in range(1, write_secret_before_test + 1):
                test_secret = {
                    "pssst": num,
                }
                self.client.secrets.kv.v2.create_or_update_secret(
                    path=path, secret=test_secret, mount_point=self.DEFAULT_MOUNT_POINT
                )
        if raises:
            with self.assertRaises(raises):
                self.client.secrets.kv.v2.read_secret_version(
                    path=path,
                    version=version,
                    mount_point=self.DEFAULT_MOUNT_POINT,
                )
        else:
            read_secret_result = self.client.secrets.kv.v2.read_secret_version(
                path=path,
                version=version,
                mount_point=self.DEFAULT_MOUNT_POINT,
            )
            logging.debug("read_secret_result: %s" % read_secret_result)
            expected_version = version or write_secret_before_test
            self.assertDictEqual(
                d1=dict(pssst=expected_version),
                d2=read_secret_result["data"]["data"],
            )

    @parameterized.expand(
        [
            ("create secret", "hvac", None, False),
            ("create secret with cas of 0", "hvac", 0, False),
            ("update secret", "hvac", None),
            ("update secret with valid cas of 1", "hvac", 1, True),
            (
                "update secret with invalid cas",
                "hvac",
                -1,
                True,
                exceptions.InvalidRequest,
                "did not match the current version",
            ),
            (
                "update with cas of 0 after path already written",
                "hvac",
                0,
                True,
                exceptions.InvalidRequest,
                "did not match the current version",
            ),
        ]
    )
    def test_create_or_update_secret(
        self,
        test_label,
        path,
        cas=None,
        write_secret_before_test=True,
        raises=None,
        exception_message="",
    ):
        test_secret = {
            "pssst": "hi",
        }

        if write_secret_before_test:
            self.client.secrets.kv.v2.create_or_update_secret(
                path=path,
                secret=test_secret,
                mount_point=self.DEFAULT_MOUNT_POINT,
            )
        if raises:
            with self.assertRaises(raises) as cm:
                self.client.secrets.kv.v2.create_or_update_secret(
                    path=path,
                    secret=test_secret,
                    cas=cas,
                    mount_point=self.DEFAULT_MOUNT_POINT,
                )
            self.assertIn(
                member=exception_message,
                container=str(cm.exception),
            )
        else:
            create_or_update_secret_result = (
                self.client.secrets.kv.v2.create_or_update_secret(
                    path=path,
                    secret=test_secret,
                    cas=cas,
                    mount_point=self.DEFAULT_MOUNT_POINT,
                )
            )
            expected_version = 2 if write_secret_before_test else 1
            logging.debug(
                "create_or_update_secret_result: %s" % create_or_update_secret_result
            )
            self.assertEqual(
                first=expected_version,
                second=create_or_update_secret_result["data"]["version"],
            )

    @parameterized.expand(
        [
            param(
                "add new key to existing secret",
                update_dict=dict(new_key="some secret"),
            ),
            param(
                "add new key to nonexistent secret",
                update_dict=dict(new_key="some secret"),
                write_secret_before_test=False,
                raises=exceptions.InvalidPath,
                exception_message="patch only works on existing data.",
            ),
            param(
                "update existing key on existing secret",
                update_dict=dict(pssst="some secret"),
            ),
        ]
    )
    def test_patch(
        self,
        label,
        update_dict,
        mount_point=DEFAULT_MOUNT_POINT,
        write_secret_before_test=True,
        raises=None,
        exception_message="",
    ):
        path = "hvac_kv_v2_test_patch"
        test_secret = {
            "pssst": "hi",
        }

        if write_secret_before_test:
            self.client.secrets.kv.v2.create_or_update_secret(
                path=path,
                secret=test_secret,
                mount_point=mount_point,
            )
        if raises:
            with self.assertRaises(raises) as cm:
                self.client.secrets.kv.v2.patch(
                    path=path,
                    secret=update_dict,
                    mount_point=mount_point,
                )
            self.assertIn(
                member=exception_message,
                container=str(cm.exception),
            )
        else:
            patch_result = self.client.secrets.kv.v2.patch(
                path=path,
                secret=update_dict,
                mount_point=mount_point,
            )
            expected_version = 2 if write_secret_before_test else 1
            logging.debug("patch_result: %s" % patch_result)
            self.assertEqual(
                first=expected_version,
                second=patch_result["data"]["version"],
            )

            read_secret_result = self.client.secrets.kv.v2.read_secret_version(
                path=path,
                mount_point=mount_point,
            )
            logging.debug("read_secret_result: %s" % read_secret_result)
            for k, v in update_dict.items():
                self.assertEqual(first=v, second=read_secret_result["data"]["data"][k])

    @parameterized.expand(
        (
            combo[0],
            combo[1],
            combo[2],
            dict(enumerate(combo)).get(3, None),
            raise_on_del,
            recoverable,
        )
        for combo in [
            ("successful delete one version written", "hvac", 1),
            ("successful delete two versions written", "hvac", 2),
            ("successful delete three versions written", "hvac", 3),
            ("nonexistent path", "no-secret-here", 0, exceptions.InvalidPath),
        ]
        for raise_on_del in [True, False]
        for recoverable in [True, False]
    )
    def test_delete_latest_version_of_secret(
        self,
        test_label,
        path,
        write_secret_before_test,
        raises=None,
        raise_on_del=False,
        recoverable=None,
        exception_message="",
    ):
        if write_secret_before_test:
            for num in range(1, write_secret_before_test + 1):
                test_secret = {
                    "pssst": num,
                }
                self.client.secrets.kv.v2.create_or_update_secret(
                    path=path, secret=test_secret, mount_point=self.DEFAULT_MOUNT_POINT
                )
        if raises:
            self.client.secrets.kv.v2.delete_latest_version_of_secret(
                path=path,
                mount_point=self.DEFAULT_MOUNT_POINT,
            )
            with self.assertRaises(raises) as cm:
                self.client.secrets.kv.v2.read_secret_metadata(
                    path=path,
                    mount_point=self.DEFAULT_MOUNT_POINT,
                )
            self.assertIn(
                member=exception_message,
                container=str(cm.exception),
            )
        else:
            delete_latest_version_of_secret_result = (
                self.client.secrets.kv.v2.delete_latest_version_of_secret(
                    path=path,
                    mount_point=self.DEFAULT_MOUNT_POINT,
                )
            )
            logging.debug(
                "delete_latest_version_of_secret_result: %s"
                % delete_latest_version_of_secret_result
            )
            read_secret_metadata_result = (
                self.client.secrets.kv.v2.read_secret_metadata(
                    path=path,
                    mount_point=self.DEFAULT_MOUNT_POINT,
                )
            )
            logging.debug(
                "read_secret_metadata_result: %s" % read_secret_metadata_result
            )
            self.assertNotEqual(
                first=read_secret_metadata_result["data"]["versions"][
                    str(write_secret_before_test)
                ]["deletion_time"],
                second="",
            )
            for num in range(1, write_secret_before_test):
                self.assertEqual(
                    first=read_secret_metadata_result["data"]["versions"][str(num)][
                        "deletion_time"
                    ],
                    second="",
                )

            should_raise = raise_on_del or not recoverable

            try:
                read_secret_version_result = (
                    self.client.secrets.kv.v2.read_secret_version(
                        path=path,
                        mount_point=self.DEFAULT_MOUNT_POINT,
                        raise_on_deleted_version=raise_on_del,
                    )
                )
            except exceptions.InvalidPath:
                if not should_raise:
                    raise
            else:
                logging.debug(
                    "read_secret_version_result: %s" % read_secret_version_result
                )
                self.assertEqual(
                    first=read_secret_version_result["data"]["data"],
                    second=None,
                )
                self.assertNotEqual(
                    first=read_secret_version_result["data"]["metadata"][
                        "deletion_time"
                    ],
                    second="",
                )
                self.assertEqual(
                    first=read_secret_version_result["data"]["metadata"]["version"],
                    second=write_secret_before_test,
                    msg=repr(read_secret_version_result),
                )

    @parameterized.expand(
        [
            ("successful delete one version written", "hvac", 1, [1]),
            ("successful delete one out of two versions written 1", "hvac", 2, [1]),
            ("successful delete one out of two versions written 2", "hvac", 2, [2]),
            ("successful delete two out of two versions written", "hvac", 2, [1, 2]),
            (
                "successful delete three out of seven versions written 135",
                "hvac",
                7,
                [1, 3, 5],
            ),
            (
                "successful delete three out of seven versions written 137",
                "hvac",
                7,
                [1, 3, 7],
            ),
            (
                "invalid versions arg none",
                "boom",
                0,
                None,
                exceptions.ParamValidationError,
            ),
            (
                "invalid versions arg empty",
                "boom",
                0,
                [],
                exceptions.ParamValidationError,
            ),
            (
                "invalid versions arg not a list",
                "goes",
                0,
                "1",
                exceptions.ParamValidationError,
            ),
            ("nonexistent version 1", "no-version-here", 3, [5]),
            ("nonexistent version 2", "no-versions-here", 3, [7, 9]),
            ("nonexistent version 3", "no-versions-here", 3, [7, 12]),
        ]
    )
    def test_delete_secret_versions(
        self,
        test_label,
        path,
        write_secret_before_test=1,
        deleted_versions=None,
        raises=None,
        exception_message="",
    ):

        if write_secret_before_test:
            for num in range(1, write_secret_before_test + 1):
                test_secret = {
                    "pssst": num,
                }
                self.client.secrets.kv.v2.create_or_update_secret(
                    path=path, secret=test_secret, mount_point=self.DEFAULT_MOUNT_POINT
                )
        if raises:
            with self.assertRaises(raises) as cm:
                self.client.secrets.kv.v2.delete_secret_versions(
                    path=path,
                    versions=deleted_versions,
                    mount_point=self.DEFAULT_MOUNT_POINT,
                )
            self.assertIn(
                member=exception_message,
                container=str(cm.exception),
            )
        else:
            delete_secret_versions_result = (
                self.client.secrets.kv.v2.delete_secret_versions(
                    path=path,
                    versions=deleted_versions,
                    mount_point=self.DEFAULT_MOUNT_POINT,
                )
            )
            logging.debug(
                "delete_secret_versions_result: %s" % delete_secret_versions_result
            )
            read_secret_metadata_result = (
                self.client.secrets.kv.v2.read_secret_metadata(
                    path=path,
                    mount_point=self.DEFAULT_MOUNT_POINT,
                )
            )
            logging.debug(
                "read_secret_metadata_result: %s" % read_secret_metadata_result
            )

            for deleted_version in deleted_versions:
                if (
                    str(deleted_version)
                    in read_secret_metadata_result["data"]["versions"]
                ):
                    self.assertNotEqual(
                        first=read_secret_metadata_result["data"]["versions"][
                            str(deleted_version)
                        ]["deletion_time"],
                        second="",
                        msg='Version "{num}" should be deleted but is not. Full read metadata response versions: {metadata}'.format(
                            num=deleted_version,
                            metadata=read_secret_metadata_result["data"]["versions"],
                        ),
                    )

            for nondeleted_version in set(range(1, write_secret_before_test + 1)) - set(
                deleted_versions
            ):
                self.assertEqual(
                    first=read_secret_metadata_result["data"]["versions"][
                        str(nondeleted_version)
                    ]["deletion_time"],
                    second="",
                    msg='Version "{num}" should not be deleted but is. Full read metadata response versions: {metadata}'.format(
                        num=nondeleted_version,
                        metadata=read_secret_metadata_result["data"]["versions"],
                    ),
                )

    @parameterized.expand(
        [
            ("successful undelete one version written", "hvac", 1, [1]),
            ("successful undelete one out of two versions written 1", "hvac", 2, [1]),
            ("successful undelete one out of two versions written 2", "hvac", 2, [2]),
            ("successful undelete two out of two versions written", "hvac", 2, [1, 2]),
            (
                "successful undelete three out of seven versions written 135",
                "hvac",
                7,
                [1, 3, 5],
            ),
            (
                "successful undelete three out of seven versions written 137",
                "hvac",
                7,
                [1, 3, 7],
            ),
            (
                "invalid versions arg none",
                "boom",
                0,
                None,
                exceptions.ParamValidationError,
            ),
            (
                "invalid versions arg empty",
                "boom",
                0,
                [],
                exceptions.ParamValidationError,
            ),
            (
                "invalid versions arg not a list",
                "goes",
                1,
                "1",
                exceptions.ParamValidationError,
            ),
            ("nonexistent version 1", "no-version-here", 3, [5]),
            ("nonexistent version 2", "no-versions-here", 3, [7, 9]),
            ("nonexistent version 3", "no-versions-here", 3, [7, 12]),
        ]
    )
    def test_undelete_secret_versions(
        self,
        test_label,
        path,
        write_secret_before_test=1,
        undeleted_versions=None,
        raises=None,
        exception_message="",
    ):

        if write_secret_before_test:
            for num in range(1, write_secret_before_test + 1):
                test_secret = {
                    "pssst": num,
                }
                self.client.secrets.kv.v2.create_or_update_secret(
                    path=path, secret=test_secret, mount_point=self.DEFAULT_MOUNT_POINT
                )
                self.client.secrets.kv.v2.delete_latest_version_of_secret(
                    path=path, mount_point=self.DEFAULT_MOUNT_POINT
                )
        if raises:
            with self.assertRaises(raises) as cm:
                self.client.secrets.kv.v2.undelete_secret_versions(
                    path=path,
                    versions=undeleted_versions,
                    mount_point=self.DEFAULT_MOUNT_POINT,
                )
            self.assertIn(
                member=exception_message,
                container=str(cm.exception),
            )
        else:
            delete_secret_versions_result = (
                self.client.secrets.kv.v2.undelete_secret_versions(
                    path=path,
                    versions=undeleted_versions,
                    mount_point=self.DEFAULT_MOUNT_POINT,
                )
            )
            logging.debug(
                "delete_secret_versions_result: %s" % delete_secret_versions_result
            )
            read_secret_metadata_result = (
                self.client.secrets.kv.v2.read_secret_metadata(
                    path=path,
                    mount_point=self.DEFAULT_MOUNT_POINT,
                )
            )
            logging.debug(
                "read_secret_metadata_result: %s" % read_secret_metadata_result
            )
            for deleted_version in undeleted_versions:
                if (
                    str(deleted_version)
                    in read_secret_metadata_result["data"]["versions"]
                ):
                    self.assertEqual(
                        first=read_secret_metadata_result["data"]["versions"][
                            str(deleted_version)
                        ]["deletion_time"],
                        second="",
                        msg='Version "{num}" should be undeleted but is not. Full read metadata response versions: {metadata}'.format(
                            num=deleted_version,
                            metadata=read_secret_metadata_result["data"]["versions"],
                        ),
                    )

            for nondeleted_version in set(range(1, write_secret_before_test + 1)) - set(
                undeleted_versions
            ):
                self.assertNotEqual(
                    first=read_secret_metadata_result["data"]["versions"][
                        str(nondeleted_version)
                    ]["deletion_time"],
                    second="",
                    msg='Version "{num}" should be deleted but it is not. Full read metadata response versions: {metadata}'.format(
                        num=nondeleted_version,
                        metadata=read_secret_metadata_result["data"]["versions"],
                    ),
                )

    @parameterized.expand(
        [
            ("successful destroy one version written", "hvac", 1, [1]),
            ("successful destroy one out of two versions written 1", "hvac", 2, [1]),
            ("successful destroy one out of two versions written 2", "hvac", 2, [2]),
            ("successful destroy two out of two versions written", "hvac", 2, [1, 2]),
            (
                "successful destroy three out of seven versions written 135",
                "hvac",
                7,
                [1, 3, 5],
            ),
            (
                "successful destroy three out of seven versions written 137",
                "hvac",
                7,
                [1, 3, 7],
            ),
            (
                "invalid versions arg None",
                "boom",
                0,
                None,
                exceptions.ParamValidationError,
            ),
            (
                "invalid versions arg empty",
                "boom",
                0,
                [],
                exceptions.ParamValidationError,
            ),
            (
                "invalid versions arg not a list",
                "goes",
                0,
                "1",
                exceptions.ParamValidationError,
            ),
            ("nonexistent version 1", "no-version-here", 3, [5]),
            ("nonexistent version 2", "no-versions-here", 3, [7, 9]),
            ("nonexistent version 3", "no-versions-here", 3, [7, 12]),
        ]
    )
    def test_destroy_secret_versions(
        self,
        test_label,
        path,
        write_secret_before_test=1,
        destoryed_versions=None,
        raises=None,
        exception_message="",
    ):

        if write_secret_before_test:
            logging.error(write_secret_before_test)
            for num in range(1, write_secret_before_test + 1):
                test_secret = {
                    "pssst": num,
                }
                self.client.secrets.kv.v2.create_or_update_secret(
                    path=path, secret=test_secret, mount_point=self.DEFAULT_MOUNT_POINT
                )
        if raises:
            with self.assertRaises(raises) as cm:
                self.client.secrets.kv.v2.destroy_secret_versions(
                    path=path,
                    versions=destoryed_versions,
                    mount_point=self.DEFAULT_MOUNT_POINT,
                )
            self.assertIn(
                member=exception_message,
                container=str(cm.exception),
            )
        else:
            destroy_secret_versions_result = (
                self.client.secrets.kv.v2.destroy_secret_versions(
                    path=path,
                    versions=destoryed_versions,
                    mount_point=self.DEFAULT_MOUNT_POINT,
                )
            )
            logging.debug(
                "destroy_secret_versions_result: %s" % destroy_secret_versions_result
            )
            read_secret_metadata_result = (
                self.client.secrets.kv.v2.read_secret_metadata(
                    path=path,
                    mount_point=self.DEFAULT_MOUNT_POINT,
                )
            )
            logging.debug(
                "read_secret_metadata_result: %s" % read_secret_metadata_result
            )
            for destoryed_version in destoryed_versions:
                if (
                    str(destoryed_version)
                    in read_secret_metadata_result["data"]["versions"]
                ):
                    self.assertTrue(
                        expr=read_secret_metadata_result["data"]["versions"][
                            str(destoryed_version)
                        ]["destroyed"],
                        msg='Version "{num}" should be destoryed but is not. Full read metadata response versions: {metadata}'.format(
                            num=destoryed_version,
                            metadata=read_secret_metadata_result["data"]["versions"],
                        ),
                    )
            for nondestoryed_version in set(
                range(1, write_secret_before_test + 1)
            ) - set(destoryed_versions):
                self.assertFalse(
                    expr=read_secret_metadata_result["data"]["versions"][
                        str(nondestoryed_version)
                    ]["destroyed"],
                    msg='Version "{num}" should not be destoryed but is. Full read metadata response versions: {metadata}'.format(
                        num=nondestoryed_version,
                        metadata=read_secret_metadata_result["data"]["versions"],
                    ),
                )

    @parameterized.expand(
        [
            (
                "nonexistent secret",
                "hvac/no-secret-here",
                False,
                exceptions.InvalidPath,
            ),
            ("list secret", "hvac/top-secret"),
        ]
    )
    def test_list_secrets(
        self,
        test_label,
        path,
        write_secret_before_test=True,
        raises=None,
        exception_message="",
    ):
        test_secret = {
            "pssst": "hi",
        }
        test_path_prefix, test_key = path.split("/")[:2]

        if write_secret_before_test:
            self.client.secrets.kv.v2.create_or_update_secret(
                path=path, secret=test_secret, mount_point=self.DEFAULT_MOUNT_POINT
            )
        if raises:
            with self.assertRaises(raises) as cm:
                self.client.secrets.kv.v2.list_secrets(
                    path=test_path_prefix, mount_point=self.DEFAULT_MOUNT_POINT
                )
            self.assertIn(
                member=exception_message,
                container=str(cm.exception),
            )
        else:
            list_secrets_result = self.client.secrets.kv.v2.list_secrets(
                path=test_path_prefix, mount_point=self.DEFAULT_MOUNT_POINT
            )

            logging.debug("list_secrets_result: %s" % list_secrets_result)
            self.assertEqual(
                first=dict(keys=[test_key]),
                second=list_secrets_result["data"],
            )

    @parameterized.expand(
        [
            ("update with no params", "hvac"),
            ("update max versions 7", "hvac", 7),
            ("update max versions 0", "hvac", 0),
            ("update cas_required true", "hvac", None, True),
            ("update cas_required false", "hvac", None, False),
            (
                "update with invalid cas_required param",
                "hvac",
                None,
                "cats",
                "0s",
                None,
                True,
                exceptions.ParamValidationError,
                "bool expected for cas_required param",
            ),
            ("update custom_medata", "hvac", None, None, "0s", dict(color="blue")),
            (
                "update with invalid custom_metadata param",
                "hvac",
                None,
                None,
                "0s",
                "not-a-dict",
                True,
                exceptions.ParamValidationError,
                "dict expected for custom_metadata param",
            ),
            ("update with delete_version_after set", "hvac", None, True, "30s"),
        ]
    )
    def test_update_metadata(
        self,
        test_label,
        path,
        max_versions=None,
        cas_required=None,
        delete_version_after="0s",
        custom_metadata=None,
        write_secret_before_test=True,
        raises=None,
        exception_message="",
    ):
        if test_label == "update custom_medata" and utils.vault_version_lt("1.9.0"):
            self.skipTest("custom_metadata support added in Vault 1.9.0")

        if write_secret_before_test:
            test_secret = {
                "pssst": "hi itsame hvac",
            }
            self.client.secrets.kv.v2.create_or_update_secret(
                path=path, secret=test_secret, mount_point=self.DEFAULT_MOUNT_POINT
            )
        if raises:
            with self.assertRaises(raises) as cm:
                self.client.secrets.kv.v2.update_metadata(
                    path=path,
                    max_versions=max_versions,
                    cas_required=cas_required,
                    custom_metadata=custom_metadata,
                    delete_version_after=delete_version_after,
                    mount_point=self.DEFAULT_MOUNT_POINT,
                )
            self.assertIn(
                member=exception_message,
                container=str(cm.exception),
            )
        else:
            update_metadata_result = self.client.secrets.kv.v2.update_metadata(
                path=path,
                max_versions=max_versions,
                cas_required=cas_required,
                custom_metadata=custom_metadata,
                delete_version_after=delete_version_after,
                mount_point=self.DEFAULT_MOUNT_POINT,
            )
            logging.debug("update_metadata_result: %s" % update_metadata_result)
            read_secret_metadata_result = (
                self.client.secrets.kv.v2.read_secret_metadata(
                    path=path,
                    mount_point=self.DEFAULT_MOUNT_POINT,
                )
            )
            logging.debug(
                "read_secret_metadata_result: %s" % read_secret_metadata_result
            )
            for key, argument in dict(
                max_versions=max_versions,
                cas_required=cas_required,
                custom_metadata=custom_metadata,
                delete_version_after=delete_version_after,
            ).items():
                if argument is not None:
                    self.assertEqual(
                        first=argument,
                        second=read_secret_metadata_result["data"][key],
                    )

    @parameterized.expand(
        [
            ("nonexistent secret", "hvac/no-secret-here", False),
            ("delete extant secret metadata", "hvac/top-secret"),
        ]
    )
    def test_delete_metadata_and_all_versions(
        self, test_label, path, write_secret_before_test=True
    ):
        test_secret = {
            "pssst": "hi",
        }

        if write_secret_before_test:
            self.client.secrets.kv.v2.create_or_update_secret(
                path=path, secret=test_secret, mount_point=self.DEFAULT_MOUNT_POINT
            )
        delete_metadata_and_all_versions_result = (
            self.client.secrets.kv.v2.delete_metadata_and_all_versions(
                path=path, mount_point=self.DEFAULT_MOUNT_POINT
            )
        )
        logging.debug(
            "delete_metadata_and_all_versions_result: %s"
            % delete_metadata_and_all_versions_result
        )
        with self.assertRaises(exceptions.InvalidPath):
            read_secret_metadata_result = (
                self.client.secrets.kv.v2.read_secret_metadata(
                    path=path,
                    mount_point=self.DEFAULT_MOUNT_POINT,
                )
            )
            logging.debug(
                "read_secret_metadata_result: %s" % read_secret_metadata_result
            )



import logging
from unittest import TestCase
from unittest import skipIf

from parameterized import parameterized, param

from hvac import exceptions
from tests import utils
from tests.utils.hvac_integration_test_case import HvacIntegrationTestCase


@skipIf(
    utils.vault_version_lt("0.10.0"),
    "Azure auth method not available before Vault version 0.10.0",
)
class TestAzure(HvacIntegrationTestCase, TestCase):
    TEST_MOUNT_POINT = "azure-test"

    def setUp(self):
        super().setUp()
        if "%s/" % self.TEST_MOUNT_POINT not in self.client.sys.list_auth_methods():
            self.client.sys.enable_auth_method(
                method_type="azure",
                path=self.TEST_MOUNT_POINT,
            )

    def tearDown(self):
        super().tearDown()
        self.client.sys.disable_auth_method(
            path=self.TEST_MOUNT_POINT,
        )

    @parameterized.expand(
        [
            param(
                "tenant_id and resource",
            ),
            param(
                "client id and secret",
                client_id="my-client-id",
                client_secret="my-client-secert",
            ),
            param(
                "invalid environment",
                environment="AzurePublicCats",
                raises=exceptions.ParamValidationError,
                exception_message="invalid environment argument provided",
            ),
        ]
    )
    def test_configure(
        self,
        label,
        client_id=None,
        client_secret=None,
        environment="AzurePublicCloud",
        raises=None,
        exception_message="",
    ):
        tenant_id = "my-tenant-id"
        resource = "my-resource"
        if raises:
            with self.assertRaises(raises) as cm:
                self.client.auth.azure.configure(
                    tenant_id=tenant_id,
                    resource=resource,
                    client_id=client_id,
                    client_secret=client_secret,
                    environment=environment,
                    mount_point=self.TEST_MOUNT_POINT,
                )
            self.assertIn(
                member=exception_message,
                container=str(cm.exception),
            )
        else:
            configure_response = self.client.auth.azure.configure(
                tenant_id=tenant_id,
                resource=resource,
                client_id=client_id,
                client_secret=client_secret,
                environment=environment,
                mount_point=self.TEST_MOUNT_POINT,
            )
            logging.debug("configure_response: %s" % configure_response)
            self.assertEqual(
                first=bool(configure_response),
                second=True,
            )

    @parameterized.expand(
        [
            param(
                "success",
            ),
            param(
                "no config written yet",
                write_config_first=False,
                raises=exceptions.InvalidPath,
            ),
        ]
    )
    def test_read_config(self, label, write_config_first=True, raises=None):
        expected_config = {
            "tenant_id": "my-tenant-id",
            "resource": "my-resource",
        }
        if write_config_first:
            self.client.auth.azure.configure(
                mount_point=self.TEST_MOUNT_POINT, **expected_config
            )
        if raises is not None:
            with self.assertRaises(raises):
                self.client.auth.azure.read_config(
                    mount_point=self.TEST_MOUNT_POINT,
                )
        else:
            read_config_response = self.client.auth.azure.read_config(
                mount_point=self.TEST_MOUNT_POINT,
            )
            logging.debug("read_config_response: %s" % read_config_response)
            for k, v in expected_config.items():
                self.assertEqual(
                    first=v,
                    second=read_config_response[k],
                )

    @parameterized.expand(
        [
            ("success",),
        ]
    )
    def test_delete_config(self, label):
        delete_config_response = self.client.auth.azure.delete_config(
            mount_point=self.TEST_MOUNT_POINT,
        )
        logging.debug("delete_config_response: %s" % delete_config_response)
        self.assertEqual(
            first=bool(delete_config_response),
            second=True,
        )

    @parameterized.expand(
        [
            param(
                "success",
                bound_service_principal_ids=["my-sp-id"],
            ),
            param(
                "CSV policies arg",
                bound_service_principal_ids=["my-sp-id"],
                policies="cats,dogs",
            ),
            param(
                "list policies arg",
                bound_service_principal_ids=["my-sp-id"],
                policies=["cats", "dogs"],
            ),
            param(
                "no bound constraints",
                raises=exceptions.InvalidRequest,
                exception_message="must have at least one bound constraint when creating/updating a role",
            ),
            param(
                "wrong policy arg type",
                bound_service_principal_ids=["my-sp-id"],
                policies={"dict": "bad"},
                raises=exceptions.ParamValidationError,
                exception_message="unsupported policies argument provided",
            ),
            param(
                "mixed policy arg type",
                bound_service_principal_ids=["my-sp-id"],
                policies=["cats", "dogs", None, 42],
                raises=exceptions.ParamValidationError,
                exception_message="unsupported policies argument provided",
            ),
        ]
    )
    def test_create_role(
        self,
        label,
        policies=None,
        bound_service_principal_ids=None,
        raises=None,
        exception_message="",
    ):
        if raises:
            with self.assertRaises(raises) as cm:
                self.client.auth.azure.create_role(
                    name="my-role",
                    policies=policies,
                    bound_service_principal_ids=bound_service_principal_ids,
                    mount_point=self.TEST_MOUNT_POINT,
                )
            self.assertIn(
                member=exception_message,
                container=str(cm.exception),
            )
        else:
            create_role_response = self.client.auth.azure.create_role(
                name="my-role",
                policies=policies,
                bound_service_principal_ids=bound_service_principal_ids,
                mount_point=self.TEST_MOUNT_POINT,
            )
            logging.debug("create_role_response: %s" % create_role_response)
            self.assertEqual(
                first=bool(create_role_response),
                second=True,
            )

    @parameterized.expand(
        [
            param(
                "success",
            ),
            param(
                "nonexistent role name",
                configure_role_first=False,
                raises=exceptions.InvalidPath,
            ),
        ]
    )
    def test_read_role(
        self,
        label,
        role_name="hvac",
        configure_role_first=True,
        raises=None,
        exception_message="",
    ):
        bound_service_principal_ids = ["some-dummy-sp-id"]
        if configure_role_first:
            create_role_response = self.client.auth.azure.create_role(
                name=role_name,
                bound_service_principal_ids=bound_service_principal_ids,
                mount_point=self.TEST_MOUNT_POINT,
            )
            logging.debug("create_role_response: %s" % create_role_response)

        if raises is not None:
            with self.assertRaises(raises):
                self.client.auth.azure.read_role(
                    name=role_name,
                    mount_point=self.TEST_MOUNT_POINT,
                )
        else:
            read_role_response = self.client.auth.azure.read_role(
                name=role_name,
                mount_point=self.TEST_MOUNT_POINT,
            )
            logging.debug("read_role_response: %s" % read_role_response)
            self.assertEqual(
                first=read_role_response["bound_service_principal_ids"],
                second=bound_service_principal_ids,
            )

    @parameterized.expand(
        [
            param(
                "success",
            ),
            param(
                "no roles",
                num_roles_to_create=0,
                raises=exceptions.InvalidPath,
            ),
            param(
                "no config",
                write_config_first=False,
            ),
        ]
    )
    def test_list_roles(
        self, label, num_roles_to_create=1, write_config_first=True, raises=None
    ):
        if write_config_first:
            self.client.auth.azure.configure(
                tenant_id="my-tenant-id",
                resource="my-resource",
                mount_point=self.TEST_MOUNT_POINT,
            )
        roles_to_create = ["hvac%s" % n for n in range(0, num_roles_to_create)]
        bound_service_principal_ids = ["some-dummy-sp-id"]
        logging.debug("roles_to_create: %s" % roles_to_create)
        for role_to_create in roles_to_create:
            create_role_response = self.client.auth.azure.create_role(
                name=role_to_create,
                bound_service_principal_ids=bound_service_principal_ids,
                mount_point=self.TEST_MOUNT_POINT,
            )
            logging.debug("create_role_response: %s" % create_role_response)

        if raises is not None:
            with self.assertRaises(raises):
                self.client.auth.azure.list_roles(
                    mount_point=self.TEST_MOUNT_POINT,
                )
        else:
            list_roles_response = self.client.auth.azure.list_roles(
                mount_point=self.TEST_MOUNT_POINT,
            )
            logging.debug("read_role_response: %s" % list_roles_response)
            self.assertEqual(
                first=list_roles_response["keys"],
                second=roles_to_create,
            )

    @parameterized.expand(
        [
            param(
                "success",
            ),
            param(
                "nonexistent role name",
                configure_role_first=False,
            ),
        ]
    )
    def test_delete_role(self, label, configure_role_first=True, raises=None):
        role_name = "hvac"
        bound_service_principal_ids = ["some-dummy-sp-id"]
        if configure_role_first:
            create_role_response = self.client.auth.azure.create_role(
                name=role_name,
                bound_service_principal_ids=bound_service_principal_ids,
                mount_point=self.TEST_MOUNT_POINT,
            )
            logging.debug("create_role_response: %s" % create_role_response)

        if raises is not None:
            with self.assertRaises(raises):
                self.client.auth.azure.delete_role(
                    name=role_name,
                    mount_point=self.TEST_MOUNT_POINT,
                )
        else:
            delete_role_response = self.client.auth.azure.delete_role(
                name=role_name,
                mount_point=self.TEST_MOUNT_POINT,
            )
            logging.debug("delete_role_response: %s" % delete_role_response)
            self.assertEqual(
                first=bool(delete_role_response),
                second=True,
            )



import json
from base64 import b64decode
from datetime import datetime
from unittest import TestCase

from unittest import mock

from hvac import Client


class TestAwsIamMethods(TestCase):
    """Unit tests providing coverage for AWS (EC2) auth backend-related methods/routes."""

    @mock.patch("hvac.aws_utils.datetime")
    @mock.patch("hvac.adapters.Adapter.login")
    def test_auth_aws_iam(self, login_mock, datetime_mock):
        datetime_mock.utcnow.return_value = datetime(2015, 8, 30, 12, 36, 0)
        client = Client()
        client.auth.aws.iam_login(
            "AKIDEXAMPLE", "wJalrXUtnFEMI/K7MDENG+bPxRfiCYEXAMPLEKEY"
        )

        login_mock.assert_called()
        args, kwargs = login_mock.call_args
        actual_params = kwargs["json"]

        actual_iam_http_request_method = actual_params["iam_http_request_method"]
        self.assertEqual("POST", actual_iam_http_request_method)

        actual_iam_request_url = b64decode(actual_params["iam_request_url"]).decode(
            "utf-8"
        )
        self.assertEqual("https://sts.amazonaws.com/", actual_iam_request_url)

        expected_auth_header_parts = [
            "Credential=AKIDEXAMPLE/20150830/us-east-1/sts/aws4_request",
            "SignedHeaders=content-length;content-type;host;x-amz-date",
            "Signature=0268ea4a725deae1116f5228d6b177fb047f9f3a9e1c5fd4baa0dc1fbb0d1a99",
        ]
        expected_iam_request_headers = {
            "Authorization": [
                "{} {}".format(
                    "AWS4-HMAC-SHA256", ", ".join(expected_auth_header_parts)
                )
            ],
            "Content-Length": ["43"],
            "Content-Type": ["application/x-www-form-urlencoded; charset=utf-8"],
            "Host": ["sts.amazonaws.com"],
            "X-Amz-Date": ["20150830T123600Z"],
        }
        actual_iam_request_headers = json.loads(
            b64decode(actual_params["iam_request_headers"])
        )
        self.assertEqual(expected_iam_request_headers, actual_iam_request_headers)

        actual_iam_request_body = b64decode(actual_params["iam_request_body"]).decode(
            "utf-8"
        )
        self.assertEqual(
            "Action=GetCallerIdentity&Version=2011-06-15", actual_iam_request_body
        )

        actual_role = actual_params["role"]
        self.assertEqual(None, actual_role)



#!/usr/bin/env python
"""Raft methods module."""
from hvac.api.system_backend.system_backend_mixin import SystemBackendMixin
from hvac import utils, adapters


class Raft(SystemBackendMixin):
    """Raft cluster-related system backend methods.

    When using Shamir seal, as soon as the Vault server is brought up, this API should be invoked
    instead of sys/init. This API completes in 2 phases. Once this is invoked, the joining node
    will receive a challenge from the Raft's leader node. This challenge can be answered by the
    joining node only after a successful unseal. Hence, the joining node should be unsealed using
    the unseal keys of the Raft's leader node.

    Reference: https://www.vaultproject.io/api-docs/system/storage/raft
    """

    def join_raft_cluster(
        self,
        leader_api_addr,
        retry=False,
        leader_ca_cert=None,
        leader_client_cert=None,
        leader_client_key=None,
    ):
        """Join a new server node to the Raft cluster.

        When using Shamir seal, as soon as the Vault server is brought up, this API should be invoked
        instead of sys/init. This API completes in 2 phases. Once this is invoked, the joining node will
        receive a challenge from the Raft's leader node. This challenge can be answered by the joining
        node only after a successful unseal. Hence, the joining node should be unsealed using the unseal
        keys of the Raft's leader node.

        Supported methods:
            POST: /sys/storage/raft/join.

        :param leader_api_addr: Address of the leader node in the Raft cluster to which this node is trying to join.
        :type leader_api_addr: str | unicode
        :param retry: Retry joining the Raft cluster in case of failures.
        :type retry: bool
        :param leader_ca_cert: CA certificate used to communicate with Raft's leader node.
        :type leader_ca_cert: str | unicode
        :param leader_client_cert: Client certificate used to communicate with Raft's leader node.
        :type leader_client_cert: str | unicode
        :param leader_client_key: Client key used to communicate with Raft's leader node.
        :type leader_client_key: str | unicode
        :return: The response of the join_raft_cluster request.
        :rtype: requests.Response
        """
        params = utils.remove_nones(
            {
                "leader_api_addr": leader_api_addr,
                "retry": retry,
                "leader_ca_cert": leader_ca_cert,
                "leader_client_cert": leader_client_cert,
                "leader_client_key": leader_client_key,
            }
        )
        api_path = "/v1/sys/storage/raft/join"
        return self._adapter.post(
            url=api_path,
            json=params,
        )

    def read_raft_config(self):
        """Read the details of all the nodes in the raft cluster.

        Supported methods:
            GET: /sys/storage/raft/configuration.

        :return: The response of the read_raft_config request.
        :rtype: requests.Response
        """
        api_path = "/v1/sys/storage/raft/configuration"
        return self._adapter.get(
            url=api_path,
        )

    def remove_raft_node(self, server_id):
        """Remove a node from the raft cluster.

        Supported methods:
            POST: /sys/storage/raft/remove-peer.

        :param server_id: The ID of the node to remove.
        :type server_id: str
        :return: The response of the remove_raft_node request.
        :rtype: requests.Response
        """
        params = {
            "server_id": server_id,
        }
        api_path = "/v1/sys/storage/raft/remove-peer"
        return self._adapter.post(
            url=api_path,
            json=params,
        )

    def take_raft_snapshot(self):
        """Returns a snapshot of the current state of the raft cluster.

        The snapshot is returned as binary data and should be redirected to a file.

        This endpoint will ignore your chosen adapter and always uses a RawAdapter.

        Supported methods:
            GET: /sys/storage/raft/snapshot.

        :return: The response of the snapshot request.
        :rtype: requests.Response
        """
        api_path = "/v1/sys/storage/raft/snapshot"
        raw_adapter = adapters.RawAdapter.from_adapter(self._adapter)
        return raw_adapter.get(
            url=api_path,
            stream=True,
        )

    def restore_raft_snapshot(self, snapshot):
        """Install the provided snapshot, returning the cluster to the state defined in it.

        Supported methods:
            POST: /sys/storage/raft/snapshot.

        :param snapshot: Previously created raft snapshot / binary data.
        :type snapshot: bytes
        :return: The response of the restore_raft_snapshot request.
        :rtype: requests.Response
        """
        api_path = "/v1/sys/storage/raft/snapshot"
        return self._adapter.post(
            url=api_path,
            data=snapshot,
        )

    def force_restore_raft_snapshot(self, snapshot):
        """Installs the provided snapshot, returning the cluster to the state defined in it.

        This is same as writing to /sys/storage/raft/snapshot except that this bypasses checks
        ensuring the Autounseal or shamir keys are consistent with the snapshot data.

        Supported methods:
            POST: /sys/storage/raft/snapshot-force.

        :param snapshot: Previously created raft snapshot / binary data.
        :type snapshot: bytes
        :return: The response of the force_restore_raft_snapshot request.
        :rtype: requests.Response
        """
        api_path = "/v1/sys/storage/raft/snapshot-force"
        return self._adapter.post(
            url=api_path,
            data=snapshot,
        )



from flask_sqlalchemy import SQLAlchemy
from authlib.integrations.sqla_oauth2 import (
    OAuth2ClientMixin,
    OAuth2TokenMixin,
    OAuth2AuthorizationCodeMixin,
)


db = SQLAlchemy()


class User(db.Model):
    id = db.Column(db.Integer, primary_key=True)
    username = db.Column(db.String(40), unique=True)

    def __str__(self):
        return self.username

    def get_user_id(self):
        return self.id

    def as_dict(self):
        return {c.name: getattr(self, c.name) for c in self.__table__.columns}


class OAuth2Client(db.Model, OAuth2ClientMixin):
    __tablename__ = "oauth2_client"

    id = db.Column(db.Integer, primary_key=True)
    user_id = db.Column(db.Integer, db.ForeignKey("user.id", ondelete="CASCADE"))
    user = db.relationship("User")

    def as_dict(self):
        return_dict = {c.name: getattr(self, c.name) for c in self.__table__.columns}
        return_dict["info"] = self.client_info
        return_dict["metadata"] = self.client_metadata
        return return_dict


class OAuth2AuthorizationCode(db.Model, OAuth2AuthorizationCodeMixin):
    __tablename__ = "oauth2_code"

    id = db.Column(db.Integer, primary_key=True)
    user_id = db.Column(db.Integer, db.ForeignKey("user.id", ondelete="CASCADE"))
    user = db.relationship("User")


class OAuth2Token(db.Model, OAuth2TokenMixin):
    __tablename__ = "oauth2_token"

    id = db.Column(db.Integer, primary_key=True)
    user_id = db.Column(db.Integer, db.ForeignKey("user.id", ondelete="CASCADE"))
    user = db.relationship("User")



from unittest import TestCase

import requests_mock
from parameterized import parameterized

from hvac.adapters import JSONAdapter
from hvac.api.auth_methods import Ldap
from hvac.api.auth_methods.ldap import DEFAULT_MOUNT_POINT


class TestLdap(TestCase):
    @parameterized.expand(
        [
            ("default mount point", DEFAULT_MOUNT_POINT),
            ("custom mount point", "other-ldap-tree"),
        ]
    )
    @requests_mock.Mocker()
    def test_configure(self, test_label, mount_point, requests_mocker):
        expected_status_code = 204
        mock_url = "http://localhost:8200/v1/auth/{mount_point}/config".format(
            mount_point=mount_point,
        )
        requests_mocker.register_uri(
            method="POST",
            url=mock_url,
            status_code=expected_status_code,
        )
        ldap = Ldap(adapter=JSONAdapter())
        response = ldap.configure(
            userdn="dc=users,cn=hvac,cn=network",
            groupdn="ou=groups,cn=hvac,cn=network",
            url="ldaps://ldap.python-hvac.org",
            mount_point=mount_point,
            anonymous_group_search=False,
            client_tls_cert="",
            client_tls_key="",
            connection_timeout=60,
            dereference_aliases="always",
            max_page_size=60,
            request_timeout=90,
            token_bound_cidrs=[],
            token_explicit_max_ttl=0,
            token_no_default_policy=False,
            token_num_uses=0,
            token_period=0,
            token_policies=[],
            token_type="",
            upndomain=None,
            userfilter="",
            username_as_alias=False,
        )
        self.assertEqual(
            first=expected_status_code,
            second=response.status_code,
        )

    @parameterized.expand(
        [
            ("default mount point", DEFAULT_MOUNT_POINT),
            ("custom mount point", "other-ldap-tree"),
        ]
    )
    @requests_mock.Mocker()
    def test_read_configuration(self, test_label, mount_point, requests_mocker):
        expected_status_code = 200
        mock_response = {
            "lease_id": "",
            "warnings": None,
            "wrap_info": None,
            "auth": None,
            "lease_duration": 0,
            "request_id": "dd7c3635-8e1c-d454-7381-bf11970fe8de",
            "data": {
                "anonymous_group_search": False,
                "binddn": "",
                "case_sensitive_names": False,
                "certificate": "",
                "connection_timeout": "",
                "deny_null_bind": True,
                "dereference_aliases": "never",
                "discoverdn": False,
                "groupattr": "",
                "groupdn": "",
                "groupfilter": "",
                "insecure_tls": False,
                "max_page_size": "0",
                "request_timeout": "90",
                "starttls": False,
                "tls_max_version": "tls12",
                "tls_min_version": "tls12",
                "upndomain": "",
                "url": "ldap://ldap.python-hvac.org",
                "userattr": "",
                "userdn": "",
            },
            "renewable": False,
        }
        mock_url = "http://localhost:8200/v1/auth/{mount_point}/config".format(
            mount_point=mount_point,
        )
        requests_mocker.register_uri(
            method="GET",
            url=mock_url,
            status_code=expected_status_code,
            json=mock_response,
        )
        ldap = Ldap(adapter=JSONAdapter())
        response = ldap.read_configuration(
            mount_point=mount_point,
        )
        self.assertEqual(
            first=mock_response,
            second=response,
        )

    @parameterized.expand(
        [
            ("default mount point", DEFAULT_MOUNT_POINT),
            ("custom mount point", "other-ldap-tree"),
        ]
    )
    @requests_mock.Mocker()
    def test_create_or_update_group(self, test_label, mount_point, requests_mocker):
        expected_status_code = 204
        group_name = "hvac"
        mock_url = (
            "http://localhost:8200/v1/auth/{mount_point}/groups/{group_name}".format(
                mount_point=mount_point,
                group_name=group_name,
            )
        )
        requests_mocker.register_uri(
            method="POST",
            url=mock_url,
            status_code=expected_status_code,
        )
        ldap = Ldap(adapter=JSONAdapter())
        response = ldap.create_or_update_group(
            name=group_name,
            mount_point=mount_point,
        )
        self.assertEqual(
            first=expected_status_code,
            second=response.status_code,
        )

    @parameterized.expand(
        [
            ("default mount point", DEFAULT_MOUNT_POINT),
            ("custom mount point", "other-ldap-tree"),
        ]
    )
    @requests_mock.Mocker()
    def test_list_groups(self, test_label, mount_point, requests_mocker):
        expected_status_code = 200
        mock_response = {
            "lease_id": "",
            "warnings": None,
            "wrap_info": None,
            "auth": None,
            "lease_duration": 0,
            "request_id": "89144def-b675-4c8a-590c-4f2ad4f1fae7",
            "data": {"keys": ["cats"]},
            "renewable": False,
        }
        mock_url = "http://localhost:8200/v1/auth/{mount_point}/groups".format(
            mount_point=mount_point,
        )
        requests_mocker.register_uri(
            method="LIST",
            url=mock_url,
            status_code=expected_status_code,
            json=mock_response,
        )
        ldap = Ldap(adapter=JSONAdapter())
        response = ldap.list_groups(
            mount_point=mount_point,
        )
        self.assertEqual(
            first=mock_response,
            second=response,
        )

    @parameterized.expand(
        [
            ("default mount point", DEFAULT_MOUNT_POINT),
            ("custom mount point", "other-ldap-tree"),
        ]
    )
    @requests_mock.Mocker()
    def test_read_group(self, test_label, mount_point, requests_mocker):
        expected_status_code = 200
        group_name = "hvac"
        mock_response = {
            "lease_id": "",
            "warnings": None,
            "wrap_info": None,
            "auth": None,
            "lease_duration": 0,
            "request_id": "448bc87c-e948-ac5f-907c-9b01fb9d26c6",
            "data": {"policies": []},
            "renewable": False,
        }
        mock_url = "http://localhost:8200/v1/auth/{mount_point}/groups/{name}".format(
            mount_point=mount_point,
            name=group_name,
        )
        requests_mocker.register_uri(
            method="GET",
            url=mock_url,
            status_code=expected_status_code,
            json=mock_response,
        )
        ldap = Ldap(adapter=JSONAdapter())
        response = ldap.read_group(
            name=group_name,
            mount_point=mount_point,
        )
        self.assertEqual(
            first=mock_response,
            second=response,
        )

    @parameterized.expand(
        [
            ("default mount point", DEFAULT_MOUNT_POINT),
            ("custom mount point", "other-ldap-tree"),
        ]
    )
    @requests_mock.Mocker()
    def test_delete_group(self, test_label, mount_point, requests_mocker):
        expected_status_code = 204
        group_name = "hvac"
        mock_url = "http://localhost:8200/v1/auth/{mount_point}/groups/{name}".format(
            mount_point=mount_point,
            name=group_name,
        )
        requests_mocker.register_uri(
            method="DELETE",
            url=mock_url,
            status_code=expected_status_code,
        )
        ldap = Ldap(adapter=JSONAdapter())
        response = ldap.delete_group(
            name=group_name,
            mount_point=mount_point,
        )
        self.assertEqual(
            first=expected_status_code,
            second=response.status_code,
        )

    @parameterized.expand(
        [
            ("default mount point", DEFAULT_MOUNT_POINT),
            ("custom mount point", "other-ldap-tree"),
        ]
    )
    @requests_mock.Mocker()
    def test_create_or_update_user(self, test_label, mount_point, requests_mocker):
        expected_status_code = 204
        username = "somedude"
        mock_url = "http://localhost:8200/v1/auth/{mount_point}/users/{name}".format(
            mount_point=mount_point,
            name=username,
        )
        requests_mocker.register_uri(
            method="POST",
            url=mock_url,
            status_code=expected_status_code,
        )
        ldap = Ldap(adapter=JSONAdapter())
        response = ldap.create_or_update_user(
            username=username,
            mount_point=mount_point,
        )
        self.assertEqual(
            first=expected_status_code,
            second=response.status_code,
        )

    @parameterized.expand(
        [
            ("default mount point", DEFAULT_MOUNT_POINT),
            ("custom mount point", "other-ldap-tree"),
        ]
    )
    @requests_mock.Mocker()
    def test_list_users(self, test_label, mount_point, requests_mocker):
        expected_status_code = 200
        mock_response = {
            "lease_id": "",
            "warnings": None,
            "wrap_info": None,
            "auth": None,
            "lease_duration": 0,
            "request_id": "0c34cc02-2f75-7deb-a531-33cf7434a729",
            "data": {"keys": ["somedude"]},
            "renewable": False,
        }
        mock_url = "http://localhost:8200/v1/auth/{mount_point}/users".format(
            mount_point=mount_point,
        )
        requests_mocker.register_uri(
            method="LIST",
            url=mock_url,
            status_code=expected_status_code,
            json=mock_response,
        )
        ldap = Ldap(adapter=JSONAdapter())
        response = ldap.list_users(
            mount_point=mount_point,
        )
        self.assertEqual(
            first=mock_response,
            second=response,
        )

    @parameterized.expand(
        [
            ("default mount point", DEFAULT_MOUNT_POINT),
            ("custom mount point", "other-ldap-tree"),
        ]
    )
    @requests_mock.Mocker()
    def test_read_user(self, test_label, mount_point, requests_mocker):
        expected_status_code = 200
        username = "somedude"
        mock_response = {
            "lease_id": "",
            "warnings": None,
            "wrap_info": None,
            "auth": None,
            "lease_duration": 0,
            "request_id": "c39914d5-70c1-b585-c6bd-ac8f0dcdf997",
            "data": {"policies": [], "groups": ""},
            "renewable": False,
        }
        mock_url = (
            "http://localhost:8200/v1/auth/{mount_point}/users/{username}".format(
                mount_point=mount_point,
                username=username,
            )
        )
        requests_mocker.register_uri(
            method="GET",
            url=mock_url,
            status_code=expected_status_code,
            json=mock_response,
        )
        ldap = Ldap(adapter=JSONAdapter())
        response = ldap.read_user(
            mount_point=mount_point,
            username=username,
        )
        self.assertEqual(
            first=mock_response,
            second=response,
        )

    @parameterized.expand(
        [
            ("default mount point", DEFAULT_MOUNT_POINT),
            ("custom mount point", "other-ldap-tree"),
        ]
    )
    @requests_mock.Mocker()
    def test_delete_user(self, test_label, mount_point, requests_mocker):
        expected_status_code = 204
        username = "somedude"
        mock_url = (
            "http://localhost:8200/v1/auth/{mount_point}/users/{username}".format(
                mount_point=mount_point,
                username=username,
            )
        )
        requests_mocker.register_uri(
            method="DELETE",
            url=mock_url,
            status_code=expected_status_code,
        )
        ldap = Ldap(adapter=JSONAdapter())
        response = ldap.delete_user(
            username=username,
            mount_point=mount_point,
        )
        self.assertEqual(
            first=expected_status_code,
            second=response.status_code,
        )

    @parameterized.expand(
        [
            ("default mount point", DEFAULT_MOUNT_POINT),
            ("custom mount point", "other-ldap-tree"),
        ]
    )
    @requests_mock.Mocker()
    def test_login(self, test_label, mount_point, requests_mocker):
        expected_status_code = 200
        username = "somedude"
        mock_response = {
            "lease_id": "",
            "warnings": None,
            "wrap_info": None,
            "auth": {
                "entity_id": "5bc030bc-2000-1176-aafb-82747ae9c874",
                "lease_duration": 2764800,
                "policies": ["default", "test-ldap-policy"],
                "client_token": "5a01125e-d823-578e-86c8-049bea022b9e",
                "accessor": "71f512de-18ab-af6e-02f7-e37b3aa48780",
                "renewable": True,
                "metadata": {"username": "somedude"},
            },
            "lease_duration": 0,
            "request_id": "c7a85e6c-fb1f-1d97-83a1-63746cb65551",
            "data": {},
            "renewable": False,
        }
        mock_url = (
            "http://localhost:8200/v1/auth/{mount_point}/login/{username}".format(
                mount_point=mount_point,
                username=username,
            )
        )
        requests_mocker.register_uri(
            method="POST",
            url=mock_url,
            status_code=expected_status_code,
            json=mock_response,
        )
        ldap = Ldap(adapter=JSONAdapter())
        response = ldap.login(
            mount_point=mount_point, username=username, password="averynicepassword"
        )
        self.assertEqual(
            first=mock_response,
            second=response,
        )



import logging
from unittest import TestCase
from unittest import skipIf

from parameterized import parameterized, param

from hvac import exceptions
from tests import utils
from tests.utils.hvac_integration_test_case import HvacIntegrationTestCase


@skipIf(
    utils.vault_version_lt("0.9.0"),
    "Identity secrets engine open sourced in Vault version >=0.9.0",
)
class TestIdentity(HvacIntegrationTestCase, TestCase):
    TEST_APPROLE_PATH = "identity-test-approle"
    TEST_APPROLE_ROLE_ID = "identity-test-role-id"
    TEST_MOUNT_POINT = "identity"
    TEST_ENTITY_NAME = "test-entity"
    TEST_ALIAS_NAME = "test-alias"
    TEST_GROUP_NAME = "test-group"
    TEST_MEMBER_GROUP_NAME = "test-group-member"
    TEST_GROUP_ALIAS_NAME = "test-group-alias"

    test_approle_accessor = None

    def setUp(self):
        super().setUp()
        if "%s/" % self.TEST_APPROLE_PATH not in self.client.sys.list_auth_methods():
            self.client.sys.enable_auth_method(
                method_type="approle",
                path=self.TEST_APPROLE_PATH,
            )
        list_auth_response = self.client.sys.list_auth_methods()
        self.test_approle_accessor = list_auth_response["data"][
            "%s/" % self.TEST_APPROLE_PATH
        ]["accessor"]

    def tearDown(self):
        super().tearDown()
        self.tear_down_entities()
        self.tear_down_entity_aliases()
        self.tear_down_groups()
        self.client.sys.disable_auth_method(
            path=self.TEST_APPROLE_PATH,
        )

    def tear_down_entities(self):
        try:
            list_entities_response = self.client.secrets.identity.list_entities(
                mount_point=self.TEST_MOUNT_POINT
            )
            logging.debug(
                "list_entities_response in tearDown: %s" % list_entities_response
            )
            entity_ids = list_entities_response["data"]["keys"]
        except exceptions.InvalidPath:
            logging.debug(
                "InvalidPath raised when calling list_entites_by_id in tearDown..."
            )
            entity_ids = []
        for entity_id in entity_ids:
            logging.debug("Deleting entity ID: %s" % entity_id)
            self.client.secrets.identity.delete_entity(
                entity_id=entity_id,
                mount_point=self.TEST_MOUNT_POINT,
            )

    def tear_down_entity_aliases(self):
        try:
            list_entity_aliases_response = (
                self.client.secrets.identity.list_entity_aliases(
                    mount_point=self.TEST_MOUNT_POINT
                )
            )
            logging.debug(
                "list_entity_aliases_response in tearDown: %s"
                % list_entity_aliases_response
            )
            alias_ids = list_entity_aliases_response["keys"]
        except exceptions.InvalidPath:
            logging.debug(
                "InvalidPath raised when calling list_entites_by_id in tearDown..."
            )
            alias_ids = []
        for alias_id in alias_ids:
            logging.debug("Deleting alias ID: %s" % alias_id)
            self.client.secrets.identity.delete_entity_alias(
                alias_id=alias_id,
                mount_point=self.TEST_MOUNT_POINT,
            )

    def tear_down_groups(self):
        try:
            list_group_response = self.client.secrets.identity.list_groups(
                mount_point=self.TEST_MOUNT_POINT
            )
            logging.debug("list_group_response in tearDown: %s" % list_group_response)
            group_ids = list_group_response["data"]["keys"]
        except exceptions.InvalidPath:
            logging.debug("InvalidPath raised when calling list_groups in tearDown...")
            group_ids = []
        for group_id in group_ids:
            logging.debug("Deleting group ID: %s" % group_id)
            self.client.secrets.identity.delete_group(
                group_id=group_id,
                mount_point=self.TEST_MOUNT_POINT,
            )

    @parameterized.expand(
        [
            param(
                "create success",
            ),
            param("create success with metadata", metadata=dict(something="meta")),
            param(
                "create failure with metadata",
                metadata="not a dict",
                raises=exceptions.ParamValidationError,
                exception_message="unsupported metadata argument provided",
            ),
            param(
                "update success",
                create_first=True,
            ),
        ]
    )
    def test_create_or_update_entity(
        self,
        label,
        metadata=None,
        create_first=False,
        raises=None,
        exception_message="",
    ):
        entity_id = None
        if create_first:
            create_first_response = (
                self.client.secrets.identity.create_or_update_entity(
                    name=self.TEST_ENTITY_NAME,
                    entity_id=entity_id,
                    metadata=metadata,
                    mount_point=self.TEST_MOUNT_POINT,
                )
            )
            logging.debug("create_first_response: %s" % create_first_response)
            entity_id = create_first_response["data"]["id"]
        if raises:
            with self.assertRaises(raises) as cm:
                self.client.secrets.identity.create_or_update_entity(
                    name=self.TEST_ENTITY_NAME,
                    metadata=metadata,
                    mount_point=self.TEST_MOUNT_POINT,
                )
            self.assertIn(
                member=exception_message,
                container=str(cm.exception),
            )
        else:
            create_or_update_response = (
                self.client.secrets.identity.create_or_update_entity(
                    name=self.TEST_ENTITY_NAME,
                    entity_id=entity_id,
                    metadata=metadata,
                    mount_point=self.TEST_MOUNT_POINT,
                )
            )
            logging.debug("create_or_update_response: %s" % create_or_update_response)
            if isinstance(create_or_update_response, dict):
                self.assertIn(
                    member="id",
                    container=create_or_update_response["data"],
                )
                if entity_id is not None:
                    self.assertEqual(
                        first=entity_id,
                        second=create_or_update_response["data"]["id"],
                    )
            else:
                self.assertEqual(
                    first=bool(create_or_update_response),
                    second=True,
                )

    @parameterized.expand(
        [
            param(
                "create success",
            ),
            param("create success with metadata", metadata=dict(something="meta")),
            param(
                "create failure with metadata",
                metadata="not a dict",
                raises=exceptions.ParamValidationError,
                exception_message="unsupported metadata argument provided",
            ),
            param(
                "update success",
                create_first=True,
            ),
        ]
    )
    @skipIf(
        utils.vault_version_lt("0.11.2"), '"by name" operations added in Vault v0.11.2'
    )
    def test_create_or_update_entity_by_name(
        self,
        label,
        metadata=None,
        create_first=False,
        raises=None,
        exception_message="",
    ):
        entity_id = None
        if create_first:
            create_first_response = (
                self.client.secrets.identity.create_or_update_entity(
                    name=self.TEST_ENTITY_NAME,
                    entity_id=entity_id,
                    metadata=metadata,
                    mount_point=self.TEST_MOUNT_POINT,
                )
            )
            logging.debug("create_first_response: %s" % create_first_response)
            entity_id = create_first_response["data"]["id"]
        if raises:
            with self.assertRaises(raises) as cm:
                self.client.secrets.identity.create_or_update_entity_by_name(
                    name=self.TEST_ENTITY_NAME,
                    metadata=metadata,
                    mount_point=self.TEST_MOUNT_POINT,
                )
            self.assertIn(
                member=exception_message,
                container=str(cm.exception),
            )
        else:
            create_or_update_response = (
                self.client.secrets.identity.create_or_update_entity_by_name(
                    name=self.TEST_ENTITY_NAME,
                    metadata=metadata,
                    mount_point=self.TEST_MOUNT_POINT,
                )
            )
            logging.debug("create_or_update_response: %s" % create_or_update_response)
            if not create_first:
                self.assertIn(
                    member="id",
                    container=create_or_update_response["data"],
                )
                if entity_id is not None:
                    self.assertEqual(
                        first=entity_id,
                        second=create_or_update_response["data"]["id"],
                    )
            else:
                self.assertEqual(
                    first=bool(create_or_update_response),
                    second=True,
                )

    @parameterized.expand(
        [
            param(
                "read success",
            ),
            param("read failure", create_first=False, raises=exceptions.InvalidPath),
        ]
    )
    def test_read_entity_by_id(
        self, label, create_first=True, raises=None, exception_message=""
    ):
        entity_id = None
        if create_first:
            create_first_response = (
                self.client.secrets.identity.create_or_update_entity(
                    name=self.TEST_ENTITY_NAME,
                    mount_point=self.TEST_MOUNT_POINT,
                )
            )
            logging.debug("create_first_response: %s" % create_first_response)
            entity_id = create_first_response["data"]["id"]
        if raises:
            with self.assertRaises(raises) as cm:
                self.client.secrets.identity.read_entity(
                    entity_id=entity_id,
                    mount_point=self.TEST_MOUNT_POINT,
                )
            self.assertIn(
                member=exception_message,
                container=str(cm.exception),
            )
        else:
            read_entity_by_id_response = self.client.secrets.identity.read_entity(
                entity_id=entity_id,
                mount_point=self.TEST_MOUNT_POINT,
            )
            logging.debug("read_entity_by_id_response: %s" % read_entity_by_id_response)
            self.assertEqual(
                first=entity_id,
                second=read_entity_by_id_response["data"]["id"],
            )

    @parameterized.expand(
        [
            param(
                "read success",
            ),
            param("read failure", create_first=False, raises=exceptions.InvalidPath),
        ]
    )
    @skipIf(
        utils.vault_version_lt("0.11.2"), '"by name" operations added in Vault v0.11.2'
    )
    def test_read_entity_by_name(
        self, label, create_first=True, raises=None, exception_message=""
    ):
        entity_id = None
        if create_first:
            create_first_response = (
                self.client.secrets.identity.create_or_update_entity(
                    name=self.TEST_ENTITY_NAME,
                    mount_point=self.TEST_MOUNT_POINT,
                )
            )
            logging.debug("create_first_response: %s" % create_first_response)
            entity_id = create_first_response["data"]["id"]
        if raises:
            with self.assertRaises(raises) as cm:
                self.client.secrets.identity.read_entity_by_name(
                    name=self.TEST_ENTITY_NAME,
                    mount_point=self.TEST_MOUNT_POINT,
                )
            self.assertIn(
                member=exception_message,
                container=str(cm.exception),
            )
        else:
            read_entity_by_name_response = (
                self.client.secrets.identity.read_entity_by_name(
                    name=self.TEST_ENTITY_NAME,
                    mount_point=self.TEST_MOUNT_POINT,
                )
            )
            logging.debug(
                "read_entity_by_name_response: %s" % read_entity_by_name_response
            )
            self.assertEqual(
                first=entity_id,
                second=read_entity_by_name_response["data"]["id"],
            )

    @parameterized.expand(
        [
            param(
                "update success",
            ),
            param("update success with metadata", metadata=dict(something="meta")),
            param(
                "update failure with metadata",
                metadata="not a dict",
                raises=exceptions.ParamValidationError,
                exception_message="unsupported metadata argument provided",
            ),
        ]
    )
    def test_update_entity(
        self, label, metadata=None, raises=None, exception_message=""
    ):
        create_first_response = self.client.secrets.identity.create_or_update_entity(
            name=self.TEST_ENTITY_NAME,
            mount_point=self.TEST_MOUNT_POINT,
        )
        logging.debug("create_first_response: %s" % create_first_response)
        entity_id = create_first_response["data"]["id"]
        if raises:
            with self.assertRaises(raises) as cm:
                self.client.secrets.identity.update_entity(
                    entity_id=entity_id,
                    metadata=metadata,
                    mount_point=self.TEST_MOUNT_POINT,
                )
            self.assertIn(
                member=exception_message,
                container=str(cm.exception),
            )
        else:
            update_entity_response = self.client.secrets.identity.update_entity(
                entity_id=entity_id,
                metadata=metadata,
                mount_point=self.TEST_MOUNT_POINT,
            )
            logging.debug("update_entity_response: %s" % update_entity_response)
            if isinstance(update_entity_response, dict):
                self.assertEqual(
                    first=update_entity_response["data"]["id"],
                    second=entity_id,
                )
            else:
                self.assertEqual(
                    first=bool(update_entity_response),
                    second=True,
                )

    @parameterized.expand(
        [
            param(
                "delete success",
            ),
            param(
                "delete success with no corresponding entity",
                create_first=False,
            ),
        ]
    )
    def test_delete_entity_by_id(
        self, label, create_first=True, raises=None, exception_message=""
    ):
        entity_id = None
        if create_first:
            create_first_response = (
                self.client.secrets.identity.create_or_update_entity(
                    name=self.TEST_ENTITY_NAME,
                    mount_point=self.TEST_MOUNT_POINT,
                )
            )
            logging.debug("create_first_response: %s" % create_first_response)
            entity_id = create_first_response["data"]["id"]
        if raises:
            with self.assertRaises(raises) as cm:
                self.client.secrets.identity.delete_entity(
                    entity_id=entity_id,
                    mount_point=self.TEST_MOUNT_POINT,
                )
            self.assertIn(
                member=exception_message,
                container=str(cm.exception),
            )
        else:
            delete_entity_response = self.client.secrets.identity.delete_entity(
                entity_id=entity_id,
                mount_point=self.TEST_MOUNT_POINT,
            )
            logging.debug("update_entity_response: %s" % delete_entity_response)
            self.assertEqual(
                first=bool(delete_entity_response),
                second=True,
            )

    @parameterized.expand(
        [
            param(
                "delete success",
            ),
            param(
                "delete success with no corresponding entity",
                create_first=False,
            ),
        ]
    )
    @skipIf(
        utils.vault_version_lt("0.11.2"), '"by name" operations added in Vault v0.11.2'
    )
    def test_delete_entity_by_name(
        self, label, create_first=True, raises=None, exception_message=""
    ):
        if create_first:
            create_first_response = (
                self.client.secrets.identity.create_or_update_entity(
                    name=self.TEST_ENTITY_NAME,
                    mount_point=self.TEST_MOUNT_POINT,
                )
            )
            logging.debug("create_first_response: %s" % create_first_response)
        if raises:
            with self.assertRaises(raises) as cm:
                self.client.secrets.identity.delete_entity_by_name(
                    name=self.TEST_ENTITY_NAME,
                    mount_point=self.TEST_MOUNT_POINT,
                )
            self.assertIn(
                member=exception_message,
                container=str(cm.exception),
            )
        else:
            delete_entity_response = self.client.secrets.identity.delete_entity_by_name(
                name=self.TEST_ENTITY_NAME,
                mount_point=self.TEST_MOUNT_POINT,
            )
            logging.debug("update_entity_response: %s" % delete_entity_response)
            self.assertEqual(
                first=bool(delete_entity_response),
                second=True,
            )

    @parameterized.expand(
        [
            param(
                "list success - LIST method",
            ),
            param(
                "list success - GET method",
                method="GET",
            ),
            param(
                "list failure - invalid method",
                method="PUT",
                raises=exceptions.ParamValidationError,
                exception_message='"method" parameter provided invalid value',
            ),
        ]
    )
    def test_list_entities_by_id(
        self, label, method="LIST", raises=None, exception_message=""
    ):
        create_response = self.client.secrets.identity.create_or_update_entity(
            name=self.TEST_ENTITY_NAME,
            mount_point=self.TEST_MOUNT_POINT,
        )
        logging.debug("create_response: %s" % create_response)
        entity_id = create_response["data"]["id"]
        if raises:
            with self.assertRaises(raises) as cm:
                self.client.secrets.identity.list_entities(
                    method=method,
                    mount_point=self.TEST_MOUNT_POINT,
                )
            self.assertIn(
                member=exception_message,
                container=str(cm.exception),
            )
        else:
            list_entities_response = self.client.secrets.identity.list_entities(
                method=method,
                mount_point=self.TEST_MOUNT_POINT,
            )
            logging.debug("list_entities_response: %s" % list_entities_response)
            self.assertEqual(
                first=[entity_id],
                second=list_entities_response["data"]["keys"],
            )

    @parameterized.expand(
        [
            param(
                "list success - LIST method",
            ),
            param(
                "list success - GET method",
                method="GET",
            ),
            param(
                "list failure - invalid method",
                method="PUT",
                raises=exceptions.ParamValidationError,
                exception_message='"method" parameter provided invalid value',
            ),
        ]
    )
    @skipIf(
        utils.vault_version_lt("0.11.2"), '"by name" operations added in Vault v0.11.2'
    )
    def test_list_entities_by_name(
        self, label, method="LIST", raises=None, exception_message=""
    ):
        create_response = self.client.secrets.identity.create_or_update_entity(
            name=self.TEST_ENTITY_NAME,
            mount_point=self.TEST_MOUNT_POINT,
        )
        logging.debug("create_response: %s" % create_response)
        if raises:
            with self.assertRaises(raises) as cm:
                self.client.secrets.identity.list_entities_by_name(
                    method=method,
                    mount_point=self.TEST_MOUNT_POINT,
                )
            self.assertIn(
                member=exception_message,
                container=str(cm.exception),
            )
        else:
            list_entities_response = self.client.secrets.identity.list_entities_by_name(
                method=method,
                mount_point=self.TEST_MOUNT_POINT,
            )
            logging.debug("list_entities_response: %s" % list_entities_response)
            self.assertEqual(
                first=[self.TEST_ENTITY_NAME],
                second=list_entities_response["data"]["keys"],
            )

    @parameterized.expand(
        [
            param(
                "merge success",
            ),
            param(
                "merge failure",
            ),
        ]
    )
    def test_merge_entities(self, label, raises=None, exception_message=""):
        create_response = self.client.secrets.identity.create_or_update_entity(
            name=self.TEST_ENTITY_NAME,
            mount_point=self.TEST_MOUNT_POINT,
        )
        logging.debug("create_response: %s" % create_response)
        create_response2 = self.client.secrets.identity.create_or_update_entity(
            name="%s2" % self.TEST_ENTITY_NAME,
            mount_point=self.TEST_MOUNT_POINT,
        )
        logging.debug("create_response2: %s" % create_response2)
        to_entity_id = create_response["data"]["id"]
        from_entity_ids = [create_response2["data"]["id"]]
        if raises:
            with self.assertRaises(raises) as cm:
                self.client.secrets.identity.merge_entities(
                    from_entity_ids=from_entity_ids,
                    to_entity_id=to_entity_id,
                    mount_point=self.TEST_MOUNT_POINT,
                )
            self.assertIn(
                member=exception_message,
                container=str(cm.exception),
            )
        else:
            merge_entities_response = self.client.secrets.identity.merge_entities(
                from_entity_ids=from_entity_ids,
                to_entity_id=to_entity_id,
                mount_point=self.TEST_MOUNT_POINT,
            )
            logging.debug("merge_entities_response: %s" % merge_entities_response)
            self.assertEqual(
                first=bool(merge_entities_response),
                second=True,
            )

    @parameterized.expand(
        [
            param(
                "merge success",
            ),
            param(
                "merge failure",
            ),
        ]
    )
    @skipIf(
        utils.vault_version_lt("1.12.0"),
        '"conflicting_alias_ids_to_keep" added in Vault v1.12.0',
    )
    def test_merge_entities_conflicting(self, label, raises=None, exception_message=""):
        create_response = self.client.secrets.identity.create_or_update_entity(
            name=self.TEST_ENTITY_NAME,
            mount_point=self.TEST_MOUNT_POINT,
        )
        logging.debug("create_response: %s" % create_response)
        create_response2 = self.client.secrets.identity.create_or_update_entity(
            name="%s2" % self.TEST_ENTITY_NAME,
            mount_point=self.TEST_MOUNT_POINT,
        )
        logging.debug("create_response2: %s" % create_response2)
        create_response3 = self.client.secrets.identity.create_or_update_entity(
            name="%s3" % self.TEST_ENTITY_NAME,
            mount_point=self.TEST_MOUNT_POINT,
        )
        logging.debug("create_response3: %s" % create_response3)
        parent_id = create_response["data"]["id"]
        merge_id1 = create_response2["data"]["id"]
        merge_id2 = create_response3["data"]["id"]

        merge_entities_response = self.client.secrets.identity.merge_entities(
            from_entity_ids=[merge_id1],
            to_entity_id=parent_id,
            mount_point=self.TEST_MOUNT_POINT,
        )
        logging.debug("merge_entities_response: %s" % merge_entities_response)

        if raises:
            with self.assertRaises(raises) as cm:
                self.client.secrets.identity.merge_entities(
                    from_entity_ids=merge_id2,
                    to_entity_id=parent_id,
                    mount_point=self.TEST_MOUNT_POINT,
                    conflicting_alias_ids_to_keep=[merge_id1],
                )
            self.assertIn(
                member=exception_message,
                container=str(cm.exception),
            )
        else:
            merge_conflicting_entities_response = (
                self.client.secrets.identity.merge_entities(
                    from_entity_ids=merge_id2,
                    to_entity_id=parent_id,
                    mount_point=self.TEST_MOUNT_POINT,
                    conflicting_alias_ids_to_keep=[merge_id1],
                )
            )
            logging.debug(
                "merge_conflicting_entities_response: %s"
                % merge_conflicting_entities_response
            )
            self.assertEqual(
                first=bool(merge_conflicting_entities_response),
                second=True,
            )

    @parameterized.expand(
        [
            param(
                "create success",
            ),
            param(
                "update success",
                create_first=True,
            ),
        ]
    )
    def test_create_or_update_entity_alias(
        self, label, create_first=False, raises=None, exception_message=""
    ):
        entity_id = None
        if create_first:
            create_first_response = (
                self.client.secrets.identity.create_or_update_entity(
                    name=self.TEST_ENTITY_NAME,
                    entity_id=entity_id,
                    mount_point=self.TEST_MOUNT_POINT,
                )
            )
            logging.debug("create_first_response: %s" % create_first_response)
            entity_id = create_first_response["data"]["id"]
        if raises:
            with self.assertRaises(raises) as cm:
                self.client.secrets.identity.create_or_update_entity_alias(
                    name=self.TEST_ALIAS_NAME,
                    canonical_id=entity_id,
                    mount_accessor=self.test_approle_accessor,
                    mount_point=self.TEST_MOUNT_POINT,
                )
            self.assertIn(
                member=exception_message,
                container=str(cm.exception),
            )
        else:
            create_or_update_response = (
                self.client.secrets.identity.create_or_update_entity_alias(
                    name=self.TEST_ALIAS_NAME,
                    canonical_id=entity_id,
                    mount_accessor=self.test_approle_accessor,
                    mount_point=self.TEST_MOUNT_POINT,
                )
            )
            logging.debug("create_or_update_response: %s" % create_or_update_response)

            self.assertIn(
                member="id",
                container=create_or_update_response["data"],
            )
            if entity_id is not None:
                self.assertEqual(
                    first=create_or_update_response["data"]["canonical_id"],
                    second=entity_id,
                )

    @parameterized.expand(
        [
            param(
                "read success",
            ),
            param(
                "read failure",
                create_first=False,
                raises=exceptions.InvalidPath,
            ),
        ]
    )
    def test_read_entity_alias_by_id(
        self, label, create_first=True, raises=None, exception_message=""
    ):
        alias_id = None
        if create_first:
            create_entity_first_response = (
                self.client.secrets.identity.create_or_update_entity(
                    name=self.TEST_ENTITY_NAME,
                    mount_point=self.TEST_MOUNT_POINT,
                )
            )
            logging.debug(
                "create_entity_first_response: %s" % create_entity_first_response
            )
            entity_id = create_entity_first_response["data"]["id"]
            create_entity_alias_first_response = (
                self.client.secrets.identity.create_or_update_entity_alias(
                    name=self.TEST_ALIAS_NAME,
                    canonical_id=entity_id,
                    mount_accessor=self.test_approle_accessor,
                    mount_point=self.TEST_MOUNT_POINT,
                )
            )
            logging.debug(
                "create_entity_alias_first_response: %s"
                % create_entity_alias_first_response
            )
            alias_id = create_entity_alias_first_response["data"]["id"]
        if raises:
            with self.assertRaises(raises) as cm:
                self.client.secrets.identity.read_entity_alias(
                    alias_id=alias_id,
                    mount_point=self.TEST_MOUNT_POINT,
                )
            self.assertIn(
                member=exception_message,
                container=str(cm.exception),
            )
        else:
            read_entity_alias_response = self.client.secrets.identity.read_entity_alias(
                alias_id=alias_id,
                mount_point=self.TEST_MOUNT_POINT,
            )
            logging.debug("read_entity_alias_response: %s" % read_entity_alias_response)

            self.assertIn(
                member="id",
                container=read_entity_alias_response["data"],
            )
            if alias_id is not None:
                self.assertEqual(
                    first=read_entity_alias_response["data"]["id"],
                    second=alias_id,
                )

    @parameterized.expand(
        [
            param(
                "update success",
            ),
            param(
                "update failure with invalid mount accessor",
                mount_accessor="not a valid accessor",
                raises=exceptions.InvalidRequest,
                exception_message="invalid mount accessor",
            ),
        ]
    )
    def test_update_entity_alias_by_id(
        self, label, mount_accessor=None, raises=None, exception_message=""
    ):
        if mount_accessor is None:
            mount_accessor = self.test_approle_accessor
        create_entity_first_response = (
            self.client.secrets.identity.create_or_update_entity(
                name=self.TEST_ENTITY_NAME,
                mount_point=self.TEST_MOUNT_POINT,
            )
        )
        logging.debug("create_entity_first_response: %s" % create_entity_first_response)
        entity_id = create_entity_first_response["data"]["id"]
        create_entity_alias_first_response = (
            self.client.secrets.identity.create_or_update_entity_alias(
                name=self.TEST_ALIAS_NAME,
                canonical_id=entity_id,
                mount_accessor=self.test_approle_accessor,
                mount_point=self.TEST_MOUNT_POINT,
            )
        )
        logging.debug(
            "create_entity_alias_first_response: %s"
            % create_entity_alias_first_response
        )
        alias_id = create_entity_alias_first_response["data"]["id"]
        if raises:
            with self.assertRaises(raises) as cm:
                self.client.secrets.identity.update_entity_alias(
                    alias_id=alias_id,
                    name=self.TEST_ALIAS_NAME,
                    canonical_id=entity_id,
                    mount_accessor=mount_accessor,
                    mount_point=self.TEST_MOUNT_POINT,
                )
            self.assertIn(
                member=exception_message,
                container=str(cm.exception),
            )
        else:
            update_entity_response = self.client.secrets.identity.update_entity_alias(
                alias_id=alias_id,
                name=self.TEST_ALIAS_NAME,
                canonical_id=entity_id,
                mount_accessor=mount_accessor,
                mount_point=self.TEST_MOUNT_POINT,
            )
            logging.debug("update_entity_response: %s" % update_entity_response)
            if isinstance(update_entity_response, dict):
                self.assertIn(
                    member="id",
                    container=update_entity_response["data"],
                )
                self.assertEqual(
                    first=update_entity_response["data"]["id"],
                    second=alias_id,
                )
            else:
                self.assertEqual(
                    first=bool(update_entity_response),
                    second=True,
                )

    @parameterized.expand(
        [
            param(
                "list success - LIST method",
            ),
            param(
                "list success - GET method",
                method="GET",
            ),
            param(
                "list failure - invalid method",
                method="PUT",
                raises=exceptions.ParamValidationError,
                exception_message='"method" parameter provided invalid value',
            ),
        ]
    )
    def test_list_entity_aliases_by_id(
        self, label, method="LIST", raises=None, exception_message=""
    ):
        create_response = self.client.secrets.identity.create_or_update_entity(
            name=self.TEST_ENTITY_NAME,
            mount_point=self.TEST_MOUNT_POINT,
        )
        logging.debug("create_response: %s" % create_response)
        entity_id = create_response["data"]["id"]
        create_entity_alias_first_response = (
            self.client.secrets.identity.create_or_update_entity_alias(
                name=self.TEST_ALIAS_NAME,
                canonical_id=entity_id,
                mount_accessor=self.test_approle_accessor,
                mount_point=self.TEST_MOUNT_POINT,
            )
        )
        alias_id = create_entity_alias_first_response["data"]["id"]
        if raises:
            with self.assertRaises(raises) as cm:
                self.client.secrets.identity.list_entity_aliases(
                    method=method,
                    mount_point=self.TEST_MOUNT_POINT,
                )
            self.assertIn(
                member=exception_message,
                container=str(cm.exception),
            )
        else:
            list_entities_response = self.client.secrets.identity.list_entity_aliases(
                method=method,
                mount_point=self.TEST_MOUNT_POINT,
            )
            logging.debug("list_entities_response: %s" % list_entities_response)
            self.assertEqual(
                first=[alias_id],
                second=list_entities_response["data"]["keys"],
            )

    @parameterized.expand(
        [
            param(
                "delete success",
            ),
            param(
                "delete success with no corresponding entity",
                create_first=False,
            ),
        ]
    )
    def test_delete_entity_alias_by_id(
        self, label, create_first=True, raises=None, exception_message=""
    ):
        alias_id = None
        if create_first:
            create_first_response = (
                self.client.secrets.identity.create_or_update_entity(
                    name=self.TEST_ENTITY_NAME,
                    mount_point=self.TEST_MOUNT_POINT,
                )
            )
            logging.debug("create_first_response: %s" % create_first_response)
            entity_id = create_first_response["data"]["id"]
            create_entity_alias_first_response = (
                self.client.secrets.identity.create_or_update_entity_alias(
                    name=self.TEST_ALIAS_NAME,
                    canonical_id=entity_id,
                    mount_accessor=self.test_approle_accessor,
                    mount_point=self.TEST_MOUNT_POINT,
                )
            )
            alias_id = create_entity_alias_first_response["data"]["id"]
        if raises:
            with self.assertRaises(raises) as cm:
                self.client.secrets.identity.delete_entity_alias(
                    alias_id=alias_id,
                    mount_point=self.TEST_MOUNT_POINT,
                )
            self.assertIn(
                member=exception_message,
                container=str(cm.exception),
            )
        else:
            delete_entity_response = self.client.secrets.identity.delete_entity_alias(
                alias_id=alias_id,
                mount_point=self.TEST_MOUNT_POINT,
            )
            logging.debug("update_entity_response: %s" % delete_entity_response)
            self.assertEqual(
                first=bool(delete_entity_response),
                second=True,
            )

    @parameterized.expand(
        [
            param(
                "create success",
            ),
            param("create success with metadata", metadata=dict(something="meta")),
            param(
                "create failure with metadata",
                metadata="not a dict",
                raises=exceptions.ParamValidationError,
                exception_message="unsupported metadata argument provided",
            ),
            param(
                "create success with group type",
                group_type="external",
                add_members=False,
            ),
            param(
                "create failure with invalid group type",
                group_type="cosmic",
                raises=exceptions.ParamValidationError,
                exception_message='unsupported group_type argument provided "cosmic"',
            ),
            param(
                "update success",
                create_first=True,
            ),
        ]
    )
    def test_create_or_update_group(
        self,
        label,
        metadata=None,
        group_type="internal",
        create_first=False,
        add_members=True,
        raises=None,
        exception_message="",
    ):
        group_id = None
        member_entity_ids = None
        member_group_ids = None
        if add_members:
            create_entity_response = (
                self.client.secrets.identity.create_or_update_entity(
                    name=self.TEST_ENTITY_NAME,
                    mount_point=self.TEST_MOUNT_POINT,
                )
            )
            logging.debug("create_entity_response: %s" % create_entity_response)
            member_entity_ids = [create_entity_response["data"]["id"]]
            create_member_group = self.client.secrets.identity.create_or_update_group(
                name=self.TEST_MEMBER_GROUP_NAME,
                mount_point=self.TEST_MOUNT_POINT,
            )
            logging.debug("create_member_group: %s" % create_member_group)
            member_group_ids = [create_member_group["data"]["id"]]
        if create_first:
            create_first_response = self.client.secrets.identity.create_or_update_group(
                name=self.TEST_GROUP_NAME,
                group_type=group_type,
                metadata=metadata,
                mount_point=self.TEST_MOUNT_POINT,
            )
            logging.debug("create_first_response: %s" % create_first_response)
            group_id = create_first_response["data"]["id"]
        if raises:
            with self.assertRaises(raises) as cm:
                self.client.secrets.identity.create_or_update_group(
                    name=self.TEST_GROUP_NAME,
                    group_id=group_id,
                    group_type=group_type,
                    metadata=metadata,
                    member_group_ids=member_group_ids,
                    member_entity_ids=member_entity_ids,
                    mount_point=self.TEST_MOUNT_POINT,
                )
            self.assertIn(
                member=exception_message,
                container=str(cm.exception),
            )
        else:
            create_or_update_response = (
                self.client.secrets.identity.create_or_update_group(
                    name=self.TEST_GROUP_NAME,
                    group_id=group_id,
                    group_type=group_type,
                    metadata=metadata,
                    member_group_ids=member_group_ids,
                    member_entity_ids=member_entity_ids,
                    mount_point=self.TEST_MOUNT_POINT,
                )
            )
            logging.debug("create_or_update_response: %s" % create_or_update_response)
            if isinstance(create_or_update_response, dict):
                self.assertIn(
                    member="id",
                    container=create_or_update_response["data"],
                )
                if group_id is not None:
                    self.assertEqual(
                        first=group_id,
                        second=create_or_update_response["data"]["id"],
                    )
                read_group_response = self.client.secrets.identity.read_group(
                    group_id=create_or_update_response["data"]["id"],
                    mount_point=self.TEST_MOUNT_POINT,
                )
                logging.debug("read_group_response: %s" % read_group_response)
                self.assertEqual(
                    first=read_group_response["data"]["member_group_ids"],
                    second=member_group_ids,
                )
                expected_member_entity_ids = (
                    member_entity_ids
                    if member_entity_ids is not None
                    else []
                    if group_type == "external"
                    and (
                        utils.vault_version_lt("1.9.9")
                    )  # https://github.com/hashicorp/vault/pull/16088
                    else None
                )
                self.assertEqual(
                    first=read_group_response["data"]["member_entity_ids"],
                    second=expected_member_entity_ids,
                )
            else:
                self.assertEqual(
                    first=bool(create_or_update_response),
                    second=True,
                )

    @parameterized.expand(
        [
            param(
                "update success",
            ),
            param("update success with metadata", metadata=dict(something="meta")),
            param(
                "update failure with metadata",
                metadata="not a dict",
                raises=exceptions.ParamValidationError,
                exception_message="unsupported metadata argument provided",
            ),
            param(
                "update failure with changed group type",
                group_type="external",
                raises=exceptions.InvalidRequest,
                exception_message="group type cannot be changed",
            ),
            param(
                "update failure with invalid group type",
                group_type="cosmic",
                raises=exceptions.ParamValidationError,
                exception_message='unsupported group_type argument provided "cosmic"',
            ),
            param(
                "update success",
                create_first=True,
            ),
        ]
    )
    def test_update_group_by_id(
        self,
        label,
        metadata=None,
        group_type="internal",
        create_first=True,
        update_members=True,
        raises=None,
        exception_message="",
    ):
        group_id = None
        member_entity_ids = None
        member_group_ids = None
        if update_members:
            create_entity_response = (
                self.client.secrets.identity.create_or_update_entity(
                    name=self.TEST_ENTITY_NAME,
                    mount_point=self.TEST_MOUNT_POINT,
                )
            )
            logging.debug("create_entity_response: %s" % create_entity_response)
            member_entity_ids = [create_entity_response["data"]["id"]]
            create_member_group = self.client.secrets.identity.create_or_update_group(
                name=self.TEST_MEMBER_GROUP_NAME,
                mount_point=self.TEST_MOUNT_POINT,
            )
            logging.debug("create_member_group: %s" % create_member_group)
            member_group_ids = [create_member_group["data"]["id"]]
        if create_first:
            create_first_response = self.client.secrets.identity.create_or_update_group(
                name=self.TEST_GROUP_NAME,
                group_type="internal",
                metadata=None,
                mount_point=self.TEST_MOUNT_POINT,
            )
            logging.debug("create_first_response: %s" % create_first_response)
            group_id = create_first_response["data"]["id"]
        if raises:
            with self.assertRaises(raises) as cm:
                self.client.secrets.identity.update_group(
                    name=self.TEST_GROUP_NAME,
                    group_id=group_id,
                    group_type=group_type,
                    metadata=metadata,
                    member_group_ids=member_group_ids,
                    member_entity_ids=member_entity_ids,
                    mount_point=self.TEST_MOUNT_POINT,
                )
            self.assertIn(
                member=exception_message,
                container=str(cm.exception),
            )
        else:
            update_response = self.client.secrets.identity.update_group(
                name=self.TEST_GROUP_NAME,
                group_id=group_id,
                group_type=group_type,
                metadata=metadata,
                member_group_ids=member_group_ids,
                member_entity_ids=member_entity_ids,
                mount_point=self.TEST_MOUNT_POINT,
            )
            logging.debug("update_response: %s" % update_response)

            if isinstance(update_response, dict):
                self.assertEqual(
                    first=update_response["data"]["id"],
                    second=group_id,
                )
            else:
                self.assertEqual(
                    first=bool(update_response),
                    second=True,
                )
            read_group_response = self.client.secrets.identity.read_group(
                group_id=group_id,
                mount_point=self.TEST_MOUNT_POINT,
            )
            logging.debug("read_group_response: %s" % read_group_response)
            self.assertEqual(
                first=read_group_response["data"]["member_group_ids"],
                second=member_group_ids,
            )
            expected_member_entity_ids = (
                member_entity_ids if member_entity_ids is not None else []
            )
            self.assertEqual(
                first=read_group_response["data"]["member_entity_ids"],
                second=expected_member_entity_ids,
            )

    @parameterized.expand(
        [
            param(
                "list success - LIST method",
            ),
            param(
                "list success - GET method",
                method="GET",
            ),
            param(
                "list failure - invalid method",
                method="PUT",
                raises=exceptions.ParamValidationError,
                exception_message='"method" parameter provided invalid value',
            ),
        ]
    )
    @skipIf(
        utils.vault_version_lt("0.11.2"), '"by name" operations added in Vault v0.11.2'
    )
    def test_list_groups_by_name(
        self, label, method="LIST", raises=None, exception_message=""
    ):
        create_response = self.client.secrets.identity.create_or_update_group(
            name=self.TEST_GROUP_NAME,
            mount_point=self.TEST_MOUNT_POINT,
        )
        logging.debug("create_response: %s" % create_response)
        if raises:
            with self.assertRaises(raises) as cm:
                self.client.secrets.identity.list_groups_by_name(
                    method=method,
                    mount_point=self.TEST_MOUNT_POINT,
                )
            self.assertIn(
                member=exception_message,
                container=str(cm.exception),
            )
        else:
            list_groups_response = self.client.secrets.identity.list_groups_by_name(
                method=method,
                mount_point=self.TEST_MOUNT_POINT,
            )
            logging.debug("list_groups_response: %s" % list_groups_response)
            self.assertEqual(
                first=[self.TEST_GROUP_NAME],
                second=list_groups_response["data"]["keys"],
            )

    @parameterized.expand(
        [
            param(
                "update success",
            ),
            param("update success with metadata", metadata=dict(something="meta")),
            param(
                "update failure with metadata",
                metadata="not a dict",
                raises=exceptions.ParamValidationError,
                exception_message="unsupported metadata argument provided",
            ),
            param(
                "update failure with changed group type",
                group_type="external",
                raises=exceptions.InvalidRequest,
                exception_message="group type cannot be changed",
            ),
            param(
                "update failure with invalid group type",
                group_type="cosmic",
                raises=exceptions.ParamValidationError,
                exception_message='unsupported group_type argument provided "cosmic"',
            ),
            param(
                "update success",
                create_first=True,
            ),
        ]
    )
    @skipIf(
        utils.vault_version_lt("0.11.2"), '"by name" operations added in Vault v0.11.2'
    )
    def test_create_or_update_group_by_name(
        self,
        label,
        metadata=None,
        group_type="internal",
        create_first=True,
        raises=None,
        exception_message="",
    ):
        if create_first:
            create_first_response = self.client.secrets.identity.create_or_update_group(
                name=self.TEST_GROUP_NAME,
                group_type="internal",
                metadata=None,
                mount_point=self.TEST_MOUNT_POINT,
            )
            logging.debug("create_first_response: %s" % create_first_response)
        if raises:
            with self.assertRaises(raises) as cm:
                self.client.secrets.identity.create_or_update_group_by_name(
                    name=self.TEST_GROUP_NAME,
                    group_type=group_type,
                    metadata=metadata,
                    mount_point=self.TEST_MOUNT_POINT,
                )
            self.assertIn(
                member=exception_message,
                container=str(cm.exception),
            )
        else:
            update_response = (
                self.client.secrets.identity.create_or_update_group_by_name(
                    name=self.TEST_GROUP_NAME,
                    group_type=group_type,
                    metadata=metadata,
                    mount_point=self.TEST_MOUNT_POINT,
                )
            )
            logging.debug("update_response: %s" % update_response)
            self.assertEqual(
                first=bool(update_response),
                second=True,
            )

    @parameterized.expand(
        [
            param(
                "read success",
            ),
            param("read failure", create_first=False, raises=exceptions.InvalidPath),
        ]
    )
    @skipIf(
        utils.vault_version_lt("0.11.2"), '"by name" operations added in Vault v0.11.2'
    )
    def test_read_group_by_name(
        self, label, create_first=True, raises=None, exception_message=""
    ):
        group_id = None
        if create_first:
            create_first_response = self.client.secrets.identity.create_or_update_group(
                name=self.TEST_GROUP_NAME,
                group_type="internal",
                metadata=None,
                mount_point=self.TEST_MOUNT_POINT,
            )
            logging.debug("create_first_response: %s" % create_first_response)
            group_id = create_first_response["data"]["id"]
        if raises:
            with self.assertRaises(raises) as cm:
                self.client.secrets.identity.read_entity_by_name(
                    name=self.TEST_GROUP_NAME,
                    mount_point=self.TEST_MOUNT_POINT,
                )
            self.assertIn(
                member=exception_message,
                container=str(cm.exception),
            )
        else:
            read_group_response = self.client.secrets.identity.read_group_by_name(
                name=self.TEST_GROUP_NAME,
                mount_point=self.TEST_MOUNT_POINT,
            )
            logging.debug("read_group_response: %s" % read_group_response)
            self.assertEqual(
                first=group_id,
                second=read_group_response["data"]["id"],
            )

    @parameterized.expand(
        [
            param(
                "create success",
            ),
            param(
                "update success",
                create_first=True,
            ),
        ]
    )
    def test_create_or_update_group_alias(
        self, label, create_first=False, raises=None, exception_message=""
    ):
        alias_id = None
        create_first_response = self.client.secrets.identity.create_or_update_group(
            name=self.TEST_ENTITY_NAME,
            mount_point=self.TEST_MOUNT_POINT,
        )
        logging.debug("create_first_response: %s" % create_first_response)
        if create_first:
            create_alias_response = (
                self.client.secrets.identity.create_or_update_group_alias(
                    name=self.TEST_GROUP_NAME,
                    mount_accessor=self.test_approle_accessor,
                    mount_point=self.TEST_MOUNT_POINT,
                )
            )
            logging.debug("create_alias_response: %s" % create_alias_response)
            alias_id = create_alias_response["data"]["id"]
        if raises:
            with self.assertRaises(raises) as cm:
                self.client.secrets.identity.create_or_update_group_alias(
                    name=self.TEST_ENTITY_NAME,
                    alias_id=alias_id,
                    mount_accessor=self.test_approle_accessor,
                    mount_point=self.TEST_MOUNT_POINT,
                )
            self.assertIn(
                member=exception_message,
                container=str(cm.exception),
            )
        else:
            create_or_update_response = (
                self.client.secrets.identity.create_or_update_group_alias(
                    name=self.TEST_GROUP_NAME,
                    alias_id=alias_id,
                    mount_accessor=self.test_approle_accessor,
                    mount_point=self.TEST_MOUNT_POINT,
                )
            )
            logging.debug("create_or_update_response: %s" % create_or_update_response)
            if "data" in create_or_update_response:
                self.assertIn(
                    member="id",
                    container=create_or_update_response["data"],
                )
                if alias_id is not None:
                    self.assertEqual(
                        first=alias_id,
                        second=create_or_update_response["data"]["id"],
                    )
            else:
                self.assertEqual(
                    first=create_or_update_response.status_code,
                    second=204,
                )

    @parameterized.expand(
        [
            param(
                "read success",
            ),
            param(
                "read failure",
                create_first=False,
                raises=exceptions.InvalidPath,
            ),
        ]
    )
    def test_read_group_alias(
        self, label, create_first=True, raises=None, exception_message=""
    ):
        alias_id = None
        create_first_response = self.client.secrets.identity.create_or_update_group(
            name=self.TEST_ENTITY_NAME,
            mount_point=self.TEST_MOUNT_POINT,
        )
        logging.debug("create_first_response: %s" % create_first_response)
        if create_first:
            create_alias_response = (
                self.client.secrets.identity.create_or_update_group_alias(
                    name=self.TEST_GROUP_NAME,
                    mount_accessor=self.test_approle_accessor,
                    mount_point=self.TEST_MOUNT_POINT,
                )
            )
            logging.debug("create_alias_response: %s" % create_alias_response)
            alias_id = create_alias_response["data"]["id"]
        if raises:
            with self.assertRaises(raises) as cm:
                self.client.secrets.identity.read_group_alias(
                    alias_id=alias_id,
                    mount_point=self.TEST_MOUNT_POINT,
                )
            self.assertIn(
                member=exception_message,
                container=str(cm.exception),
            )
        else:
            read_group_alias_response = self.client.secrets.identity.read_group_alias(
                alias_id=alias_id,
                mount_point=self.TEST_MOUNT_POINT,
            )
            logging.debug("read_group_alias_response: %s" % read_group_alias_response)

            self.assertIn(
                member="id",
                container=read_group_alias_response["data"],
            )
            if alias_id is not None:
                self.assertEqual(
                    first=read_group_alias_response["data"]["id"],
                    second=alias_id,
                )

    @parameterized.expand(
        [
            param(
                "list success - LIST method",
            ),
            param(
                "list success - GET method",
                method="GET",
            ),
            param(
                "list failure - invalid method",
                method="PUT",
                raises=exceptions.ParamValidationError,
                exception_message='"method" parameter provided invalid value',
            ),
        ]
    )
    def test_list_group_aliases(
        self, label, method="LIST", raises=None, exception_message=""
    ):
        create_group_response = self.client.secrets.identity.create_or_update_group(
            name=self.TEST_GROUP_ALIAS_NAME,
            group_type="internal",
            metadata=None,
            mount_point=self.TEST_MOUNT_POINT,
        )
        logging.debug("create_group_response: %s" % create_group_response)
        create_alias_response = (
            self.client.secrets.identity.create_or_update_group_alias(
                name=self.TEST_GROUP_NAME,
                mount_accessor=self.test_approle_accessor,
                mount_point=self.TEST_MOUNT_POINT,
            )
        )
        logging.debug("create_alias_response: %s" % create_alias_response)
        alias_id = create_alias_response["data"]["id"]
        if raises:
            with self.assertRaises(raises) as cm:
                self.client.secrets.identity.list_group_aliases(
                    method=method,
                    mount_point=self.TEST_MOUNT_POINT,
                )
            self.assertIn(
                member=exception_message,
                container=str(cm.exception),
            )
        else:
            list_groups_response = self.client.secrets.identity.list_group_aliases(
                method=method,
                mount_point=self.TEST_MOUNT_POINT,
            )
            logging.debug("list_groups_response: %s" % list_groups_response)
            self.assertEqual(
                first=[alias_id],
                second=list_groups_response["data"]["keys"],
            )

    @parameterized.expand(
        [
            param(
                "lookup entity",
                criteria=["entity_id"],
            ),
            param(
                "lookup entity alias",
                criteria=["alias_id"],
            ),
            param(
                "lookup missing entity",
                criteria=["entity_id"],
                create_first=False,
            ),
        ]
    )
    def test_lookup_entity(
        self, label, criteria, create_first=True, raises=None, exception_message=""
    ):
        lookup_params = {}
        if create_first:
            create_entity_response = (
                self.client.secrets.identity.create_or_update_entity(
                    name=self.TEST_ENTITY_NAME,
                    mount_point=self.TEST_MOUNT_POINT,
                )
            )
            logging.debug("create_entity_response: %s" % create_entity_response)
            entity_id = create_entity_response["data"]["id"]
            create_alias_response = (
                self.client.secrets.identity.create_or_update_entity_alias(
                    name=self.TEST_ALIAS_NAME,
                    canonical_id=entity_id,
                    mount_accessor=self.test_approle_accessor,
                    mount_point=self.TEST_MOUNT_POINT,
                )
            )
            logging.debug("create_alias_response: %s" % create_alias_response)
            alias_id = create_alias_response["data"]["id"]
            if "entity_id" in criteria:
                lookup_params["entity_id"] = entity_id
            elif "alias_id" in criteria:
                lookup_params["alias_id"] = alias_id
        else:
            for key in criteria:
                lookup_params[key] = key
        logging.debug("lookup_params: %s" % lookup_params)
        if raises:
            with self.assertRaises(raises) as cm:
                self.client.secrets.identity.lookup_entity(
                    mount_point=self.TEST_MOUNT_POINT, **lookup_params
                )
            self.assertIn(
                member=exception_message,
                container=str(cm.exception),
            )
        else:
            lookup_entity_response = self.client.secrets.identity.lookup_entity(
                mount_point=self.TEST_MOUNT_POINT, **lookup_params
            )
            logging.debug("lookup_entity_response: %s" % lookup_entity_response)
            if create_first:
                if "entity_id" in criteria:
                    self.assertEqual(
                        first=lookup_entity_response["data"]["name"],
                        second=self.TEST_ENTITY_NAME,
                    )
                elif "alias_id" in criteria:
                    self.assertEqual(
                        first=lookup_entity_response["data"]["aliases"][0]["name"],
                        second=self.TEST_ALIAS_NAME,
                    )
            else:
                self.assertEqual(
                    first=lookup_entity_response.status_code,
                    second=204,
                )

    @parameterized.expand(
        [
            param(
                "lookup group",
                criteria=["group_id"],
            ),
            param(
                "lookup group alias",
                criteria=["alias_id"],
            ),
            param(
                "lookup name",
                criteria=["name"],
            ),
            param(
                "lookup alias",
                criteria=["alias_name", "alias_mount_accessor"],
            ),
            param(
                "lookup missing group",
                criteria=["group_id"],
                create_first=False,
            ),
        ]
    )
    def test_lookup_group(
        self, label, criteria, create_first=True, raises=None, exception_message=""
    ):
        lookup_params = {}
        if create_first:
            create_group_response = self.client.secrets.identity.create_or_update_group(
                name=self.TEST_GROUP_NAME,
                group_type="external",
                mount_point=self.TEST_MOUNT_POINT,
            )
            logging.debug("create_group_response: %s" % create_group_response)
            group_id = create_group_response["data"]["id"]
            create_alias_response = (
                self.client.secrets.identity.create_or_update_group_alias(
                    name=self.TEST_GROUP_ALIAS_NAME,
                    canonical_id=group_id,
                    mount_accessor=self.test_approle_accessor,
                    mount_point=self.TEST_MOUNT_POINT,
                )
            )
            logging.debug("create_alias_response: %s" % create_alias_response)
            alias_id = create_alias_response["data"]["id"]
            if "group_id" in criteria:
                lookup_params["group_id"] = group_id
            elif "alias_id" in criteria:
                lookup_params["alias_id"] = alias_id
            elif "name" in criteria:
                lookup_params["name"] = self.TEST_GROUP_NAME
            elif "alias_name" in criteria and "alias_mount_accessor" in criteria:
                lookup_params["alias_name"] = self.TEST_GROUP_ALIAS_NAME
                lookup_params["alias_mount_accessor"] = self.test_approle_accessor
        else:
            for key in criteria:
                lookup_params[key] = key
        logging.debug("lookup_params: %s" % lookup_params)
        if raises:
            with self.assertRaises(raises) as cm:
                self.client.secrets.identity.lookup_group(
                    mount_point=self.TEST_MOUNT_POINT, **lookup_params
                )
            self.assertIn(
                member=exception_message,
                container=str(cm.exception),
            )
        else:
            lookup_group_response = self.client.secrets.identity.lookup_group(
                mount_point=self.TEST_MOUNT_POINT, **lookup_params
            )
            logging.debug("lookup_group_response: %s" % lookup_group_response)
            if create_first:
                if "group_id" in criteria or "name" in criteria:
                    self.assertEqual(
                        first=lookup_group_response["data"]["name"],
                        second=self.TEST_GROUP_NAME,
                    )
                elif "alias_id" in criteria or (
                    "alias_name" in criteria and "alias_mount_accessor" in criteria
                ):
                    self.assertEqual(
                        first=lookup_group_response["data"]["alias"]["name"],
                        second=self.TEST_GROUP_ALIAS_NAME,
                    )
            else:
                self.assertEqual(
                    first=lookup_group_response.status_code,
                    second=204,
                )

    @parameterized.expand(
        [
            param(
                "empty issuer",
                issuer="",
            ),
            param(
                "issuer set",
                issuer="https://python-hvac.org:1234",
            ),
        ]
    )
    def test_configure_tokens_backend(self, label, issuer):
        response = self.client.secrets.identity.configure_tokens_backend(
            issuer=issuer,
        )
        logging.debug("configure_tokens_backend response: %s" % response)
        if issuer:
            # e.g.: 'warnings': ['If "issuer" is set explicitly, [...]']
            self.assertGreaterEqual(
                a=len(response["warnings"]),
                b=1,
            )
        else:
            # No response body if we're _not_ setting the issue
            self.assertEqual(
                first=204,
                second=response.status_code,
            )

    @parameterized.expand(
        [
            param(
                "empty issuer",
                issuer="",
            ),
            param(
                "issuer set",
                issuer="",
            ),
        ]
    )
    def test_read_tokens_backend_configuration(self, label, issuer):
        configure_tokens_backend_response = (
            self.client.secrets.identity.configure_tokens_backend(
                issuer=issuer,
            )
        )
        logging.debug(
            "configure_tokens_backend_response: %s" % configure_tokens_backend_response
        )
        response = self.client.secrets.identity.read_tokens_backend_configuration()
        logging.debug("read_tokens_backend_configuration response: %s" % response)
        self.assertEqual(
            first=issuer,
            second=response["data"]["issuer"],
        )

    @parameterized.expand(
        [
            param(
                "name set",
                name="hvac",
            ),
        ]
    )
    def test_create_named_key(self, label, name):
        response = self.client.secrets.identity.create_named_key(
            name=name,
        )
        logging.debug("create_named_key response: %s" % response)
        self.assertEqual(
            first=204,
            second=response.status_code,
        )

    @parameterized.expand(
        [
            param(
                "success",
                name="hvac",
                algorithm="ES256",
            ),
        ]
    )
    def test_read_named_key(self, label, name, algorithm):
        create_named_key_response = self.client.secrets.identity.create_named_key(
            name=name,
            algorithm=algorithm,
        )
        logging.debug("create_named_key response: %s" % create_named_key_response)
        response = self.client.secrets.identity.read_named_key(
            name=name,
        )
        logging.debug("read_named_key response: %s" % response)
        self.assertEqual(
            first=algorithm,
            second=response["data"]["algorithm"],
        )

    @parameterized.expand(
        [
            param(
                "success",
                name="hvac",
            ),
        ]
    )
    def test_delete_named_key(self, label, name):
        create_named_key_response = self.client.secrets.identity.create_named_key(
            name=name,
        )
        logging.debug("create_named_key response: %s" % create_named_key_response)
        response = self.client.secrets.identity.delete_named_key(
            name=name,
        )
        logging.debug("delete_named_key response: %s" % response)
        self.assertEqual(
            first=204,
            second=response.status_code,
        )

    @parameterized.expand(
        [
            param(
                "success",
                name="hvac",
            ),
        ]
    )
    def test_list_named_keys(self, label, name):
        create_named_key_response = self.client.secrets.identity.create_named_key(
            name=name,
        )
        logging.debug("create_named_key response: %s" % create_named_key_response)
        response = self.client.secrets.identity.list_named_keys()
        logging.debug("list_named_keys response: %s" % response)
        self.assertIn(
            member=name,
            container=response["data"]["keys"],
        )

    @parameterized.expand(
        [
            param(
                "lower ttl than at create time",
                name="hvac",
                verification_ttl=1800,
            ),
        ]
    )
    def test_rotate_named_key(self, label, name, verification_ttl):
        create_named_key_response = self.client.secrets.identity.create_named_key(
            name=name,
            verification_ttl=verification_ttl + 1,
        )
        logging.debug("create_named_key response: %s" % create_named_key_response)
        response = self.client.secrets.identity.rotate_named_key(
            name=name,
            verification_ttl=verification_ttl,
        )
        logging.debug("rotate_named_key response: %s" % response)
        self.assertEqual(
            first=204,
            second=response.status_code,
        )
        post_rotate_read_response = self.client.secrets.identity.read_named_key(
            name=name,
        )
        logging.debug(
            "post_rotate_read_response response: %s" % post_rotate_read_response
        )
        self.assertEqual(
            first=verification_ttl,
            second=post_rotate_read_response["data"]["verification_ttl"],
        )

    @parameterized.expand(
        [
            param(
                "success",
                name="hvac",
                key_name="hvac_key",
            ),
        ]
    )
    def test_create_or_update_role(self, label, name, key_name):
        create_named_key_response = self.client.secrets.identity.create_named_key(
            name=key_name,
        )
        logging.debug("create_named_key response: %s" % create_named_key_response)
        response = self.client.secrets.identity.create_or_update_role(
            name=name,
            key=key_name,
        )
        logging.debug("create_or_update_role response: %s" % response)
        self.assertEqual(
            first=204,
            second=response.status_code,
        )

    @parameterized.expand(
        [
            param(
                "success",
                name="hvac",
                key_name="hvac_key",
            ),
        ]
    )
    def test_read_role(self, label, name, key_name):
        create_named_key_response = self.client.secrets.identity.create_named_key(
            name=key_name,
        )
        logging.debug("create_named_key response: %s" % create_named_key_response)
        create_or_update_role_response = (
            self.client.secrets.identity.create_or_update_role(
                name=name,
                key=key_name,
            )
        )
        logging.debug(
            "create_or_update_role response: %s" % create_or_update_role_response
        )
        response = self.client.secrets.identity.read_role(
            name=name,
        )
        logging.debug("read_role response: %s" % response)
        self.assertEqual(
            first=key_name,
            second=response["data"]["key"],
        )

    @parameterized.expand(
        [
            param(
                "success",
                name="hvac",
                key_name="hvac_key",
            ),
        ]
    )
    def test_delete_role(self, label, name, key_name):
        create_named_key_response = self.client.secrets.identity.create_named_key(
            name=key_name,
        )
        logging.debug("create_named_key response: %s" % create_named_key_response)
        create_or_update_role_response = (
            self.client.secrets.identity.create_or_update_role(
                name=name,
                key=key_name,
            )
        )
        logging.debug(
            "create_or_update_role response: %s" % create_or_update_role_response
        )
        response = self.client.secrets.identity.delete_role(
            name=name,
        )
        logging.debug("delete_role response: %s" % response)
        self.assertEqual(
            first=204,
            second=response.status_code,
        )

    @parameterized.expand(
        [
            param(
                "success",
                name="hvac",
                key_name="hvac_key",
            ),
        ]
    )
    def test_list_roles(self, label, name, key_name):
        create_named_key_response = self.client.secrets.identity.create_named_key(
            name=key_name,
        )
        logging.debug("create_named_key response: %s" % create_named_key_response)
        create_or_update_role_response = (
            self.client.secrets.identity.create_or_update_role(
                name=name,
                key=key_name,
            )
        )
        logging.debug(
            "create_or_update_role response: %s" % create_or_update_role_response
        )
        response = self.client.secrets.identity.list_roles()
        logging.debug("list_roles response: %s" % response)
        self.assertIn(
            member=name,
            container=response["data"]["keys"],
        )

    @parameterized.expand(
        [
            param(
                "success",
                role_name="hvac",
                key_name="hvac_key",
            ),
        ]
    )
    def test_generate_signed_id_token(self, label, role_name, key_name):
        create_or_update_role_response = (
            self.client.secrets.identity.create_or_update_role(
                name=role_name,
                key=key_name,
            )
        )
        logging.debug(
            "create_or_update_role response: %s" % create_or_update_role_response
        )
        read_role_response = self.client.secrets.identity.read_role(
            name=role_name,
        )
        logging.debug("read_role response: %s" % read_role_response)
        token_client_id = read_role_response["data"]["client_id"]
        create_named_key_response = self.client.secrets.identity.create_named_key(
            name=key_name,
            allowed_client_ids=[
                token_client_id,
            ],
        )
        logging.debug("create_named_key response: %s" % create_named_key_response)
        # Log in using a dummy approle role so our client token has an associated identity
        self.login_using_admin_approle_role(
            role_id=self.TEST_APPROLE_ROLE_ID,
            path=self.TEST_APPROLE_PATH,
        )
        response = self.client.secrets.identity.generate_signed_id_token(
            name=role_name,
        )
        logging.debug("generate_signed_id_token response: %s" % response)
        self.assertIn(
            member=token_client_id,
            container=response["data"]["client_id"],
        )

    @parameterized.expand(
        [
            param(
                "success",
                role_name="hvac",
                key_name="hvac_key",
            ),
        ]
    )
    def test_introspect_signed_id_token(self, label, role_name, key_name):
        create_or_update_role_response = (
            self.client.secrets.identity.create_or_update_role(
                name=role_name,
                key=key_name,
            )
        )
        logging.debug(
            "create_or_update_role response: %s" % create_or_update_role_response
        )
        read_role_response = self.client.secrets.identity.read_role(
            name=role_name,
        )
        logging.debug("read_role response: %s" % read_role_response)
        token_client_id = read_role_response["data"]["client_id"]
        create_named_key_response = self.client.secrets.identity.create_named_key(
            name=key_name,
            allowed_client_ids=[
                token_client_id,
            ],
        )
        logging.debug("create_named_key response: %s" % create_named_key_response)
        # Log in using a dummy approle role so our client token has an associated identity
        self.login_using_admin_approle_role(
            role_id=self.TEST_APPROLE_ROLE_ID,
            path=self.TEST_APPROLE_PATH,
        )
        generate_signed_id_token_response = (
            self.client.secrets.identity.generate_signed_id_token(
                name=role_name,
            )
        )
        logging.debug(
            "generate_signed_id_token response: %s" % generate_signed_id_token_response
        )
        response = self.client.secrets.identity.introspect_signed_id_token(
            token=generate_signed_id_token_response["data"]["token"],
            client_id=token_client_id,
        )
        logging.debug("introspect_signed_id_token response: %s" % response)
        self.assertIn(
            member="active",
            container=response,
        )
        self.assertTrue(
            expr=response["active"],
        )

    @parameterized.expand(
        [
            param(
                "issuer set",
                issuer="https://python-hvac.org:1234",
            ),
        ]
    )
    def test_read_well_known_configurations(self, label, issuer):
        response = self.client.secrets.identity.configure_tokens_backend(
            issuer=issuer,
        )
        response = self.client.secrets.identity.read_well_known_configurations()
        logging.debug("read_well_known_configurations response: %s" % response)
        self.assertIn(
            member=issuer,
            container=response["issuer"],
        )

    @parameterized.expand(
        [
            param(
                "success",
                issuer="https://python-hvac.org:1234",
            ),
        ]
    )
    def test_read_active_public_keys(self, label, issuer):
        response = self.client.secrets.identity.configure_tokens_backend(
            issuer=issuer,
        )
        response = self.client.secrets.identity.read_active_public_keys()
        logging.debug("read_active_public_keys response: %s" % response)
        self.assertIn(
            member="keys",
            container=response,
        )



#!/usr/bin/env python
"""Multi-factor authentication methods module."""
from hvac.api.auth_methods.legacy_mfa import LegacyMfa
from hvac.api.vault_api_base import VaultApiBase
from hvac import exceptions, utils

SUPPORTED_MFA_TYPES = [
    "duo",
]
SUPPORTED_AUTH_METHODS = ["ldap", "okta", "radius", "userpass"]


class Mfa(VaultApiBase):
    """Multi-factor authentication Auth Method (API).

    .. warning::
        This class's methods correspond to a legacy / unsupported set of Vault API routes. Please see the reference link
        for additional context.

    Reference: https://www.vaultproject.io/docs/auth/mfa.html
    """

    @utils.deprecated_method(
        to_be_removed_in_version="2.0.0",
        new_method=LegacyMfa.configure,
    )
    def configure(self, mount_point, mfa_type="duo", force=False):
        """Configure MFA for a supported method.

        This endpoint allows you to turn on multi-factor authentication with a given backend.
        Currently only Duo is supported.

        Supported methods:
            POST: /auth/{mount_point}/mfa_config. Produces: 204 (empty body)

        :param mount_point: The "path" the method/backend was mounted on.
        :type mount_point: str | unicode
        :param mfa_type: Enables MFA with given backend (available: duo)
        :type mfa_type: str | unicode
        :param force: If True, make the "mfa_config" request regardless of circumstance. If False (the default), verify
            the provided mount_point is available and one of the types of methods supported by this feature.
        :type force: bool
        :return: The response of the configure MFA request.
        :rtype: requests.Response
        """
        if mfa_type != "duo" and not force:
            # The situation described via this exception is not likely to change in the future.
            # However we provided that flexibility here just in case.
            error_msg = 'Unsupported mfa_type argument provided "{arg}", supported types: "{mfa_types}"'
            raise exceptions.ParamValidationError(
                error_msg.format(
                    mfa_types=",".join(SUPPORTED_MFA_TYPES),
                    arg=mfa_type,
                )
            )
        params = {
            "type": mfa_type,
        }

        api_path = utils.format_url(
            "/v1/auth/{mount_point}/mfa_config", mount_point=mount_point
        )
        return self._adapter.post(
            url=api_path,
            json=params,
        )

    @utils.deprecated_method(
        to_be_removed_in_version="2.0.0",
        new_method=LegacyMfa.read_configuration,
    )
    def read_configuration(self, mount_point):
        """Read the MFA configuration.

        Supported methods:
            GET: /auth/{mount_point}/mfa_config. Produces: 200 application/json


        :param mount_point: The "path" the method/backend was mounted on.
        :type mount_point: str | unicode
        :return: The JSON response of the read_configuration request.
        :rtype: dict
        """
        api_path = utils.format_url(
            "/v1/auth/{mount_point}/mfa_config",
            mount_point=mount_point,
        )
        return self._adapter.get(url=api_path)

    @utils.deprecated_method(
        to_be_removed_in_version="2.0.0",
        new_method=LegacyMfa.configure_duo_access,
    )
    def configure_duo_access(self, mount_point, host, integration_key, secret_key):
        """Configure the access keys and host for Duo API connections.

        To authenticate users with Duo, the backend needs to know what host to connect to and must authenticate with an
        integration key and secret key. This endpoint is used to configure that information.

        Supported methods:
            POST: /auth/{mount_point}/duo/access. Produces: 204 (empty body)

        :param mount_point: The "path" the method/backend was mounted on.
        :type mount_point: str | unicode
        :param host: Duo API host
        :type host: str | unicode
        :param integration_key: Duo integration key
        :type integration_key: Duo secret key
        :param secret_key: The "path" the method/backend was mounted on.
        :type secret_key: str | unicode
        :return: The response of the configure_duo_access request.
        :rtype: requests.Response
        """
        params = {
            "host": host,
            "ikey": integration_key,
            "skey": secret_key,
        }
        api_path = utils.format_url(
            "/v1/auth/{mount_point}/duo/access",
            mount_point=mount_point,
        )
        return self._adapter.post(
            url=api_path,
            json=params,
        )

    @utils.deprecated_method(
        to_be_removed_in_version="2.0.0",
        new_method=LegacyMfa.configure_duo_behavior,
    )
    def configure_duo_behavior(
        self, mount_point, push_info=None, user_agent=None, username_format="%s"
    ):
        """Configure Duo second factor behavior.

        This endpoint allows you to configure how the original auth method username maps to the Duo username by
        providing a template format string.

        Supported methods:
            POST: /auth/{mount_point}/duo/config. Produces: 204 (empty body)


        :param mount_point: The "path" the method/backend was mounted on.
        :type mount_point: str | unicode
        :param push_info: A string of URL-encoded key/value pairs that provides additional context about the
            authentication attempt in the Duo Mobile app
        :type push_info: str | unicode
        :param user_agent: User agent to connect to Duo (default "")
        :type user_agent: str | unicode
        :param username_format: Format string given auth method username as argument to create Duo username
            (default '%s')
        :type username_format: str | unicode
        :return: The response of the configure_duo_behavior request.
        :rtype: requests.Response
        """
        params = {
            "username_format": username_format,
        }
        if push_info is not None:
            params["push_info"] = push_info
        if user_agent is not None:
            params["user_agent"] = user_agent
        api_path = utils.format_url(
            "/v1/auth/{mount_point}/duo/config",
            mount_point=mount_point,
        )
        return self._adapter.post(
            url=api_path,
            json=params,
        )

    @utils.deprecated_method(
        to_be_removed_in_version="2.0.0",
        new_method=LegacyMfa.read_duo_behavior_configuration,
    )
    def read_duo_behavior_configuration(self, mount_point):
        """Read the Duo second factor behavior configuration.

        Supported methods:
            GET: /auth/{mount_point}/duo/config. Produces: 200 application/json


        :param mount_point: The "path" the method/backend was mounted on.
        :type mount_point: str | unicode
        :return: The JSON response of the read_duo_behavior_configuration request.
        :rtype: dict
        """
        api_path = utils.format_url(
            "/v1/auth/{mount_point}/duo/config",
            mount_point=mount_point,
        )
        return self._adapter.get(url=api_path)



from unittest import TestCase

from hvac import exceptions
from parameterized import parameterized
from tests import utils
from tests.utils.hvac_integration_test_case import HvacIntegrationTestCase


class TestCert(HvacIntegrationTestCase, TestCase):
    TEST_MOUNT_POINT = "cert-test"
    TEST_ROLE_NAME = "testrole"
    TEST_CLIENT_CERTIFICATE_FILE = utils.get_config_file_path("client-cert.pem")
    cert = utils.create_client()._adapter._kwargs.get("cert")
    with open(TEST_CLIENT_CERTIFICATE_FILE, "r") as fp:
        TEST_CERTIFICATE = fp.read()

    def setUp(self):
        super().setUp()
        if "%s/" % self.TEST_MOUNT_POINT not in self.client.sys.list_auth_methods():
            self.client.sys.enable_auth_method(
                method_type="cert",
                path=self.TEST_MOUNT_POINT,
            )
        _ = self.client.auth.cert.create_ca_certificate_role(
            name=self.TEST_ROLE_NAME,
            certificate=self.TEST_CERTIFICATE,
            mount_point=self.TEST_MOUNT_POINT,
        )

    def tearDown(self):
        super().tearDown()

    def test_create_ca_certificate_role(self):
        response = self.client.auth.cert.create_ca_certificate_role(
            name="testrole2",
            certificate=self.TEST_CERTIFICATE,
            mount_point=self.TEST_MOUNT_POINT,
        )

        self.assertEqual(first=204, second=response.status_code)

    def test_create_ca_certificate_with_filename(self):
        response = self.client.auth.cert.create_ca_certificate_role(
            name="testrole2",
            certificate_file=self.TEST_CLIENT_CERTIFICATE_FILE,
            mount_point=self.TEST_MOUNT_POINT,
        )

        self.assertEqual(first=204, second=response.status_code)

    def test_create_ca_certificate_with_filename_deprecated(self):
        """This tests the deprecated feature of passing a certificate file via the certificate argument"""
        response = self.client.auth.cert.create_ca_certificate_role(
            name="testrole2",
            certificate=self.TEST_CLIENT_CERTIFICATE_FILE,
            mount_point=self.TEST_MOUNT_POINT,
        )

        self.assertEqual(first=204, second=response.status_code)

    def test_read_ca_certificate_role(self):
        response = self.client.auth.cert.read_ca_certificate_role(
            name=self.TEST_ROLE_NAME,
            mount_point=self.TEST_MOUNT_POINT,
        )

        self.assertEqual(
            first=self.TEST_ROLE_NAME,
            second=response["data"]["display_name"],
        )

    def test_list_certificate_roles(self):
        response = self.client.auth.cert.list_certificate_roles(
            mount_point=self.TEST_MOUNT_POINT,
        )

        self.assertEqual(first=response["data"]["keys"], second=[self.TEST_ROLE_NAME])

    def test_delete_certificate_role(self):
        self.client.auth.cert.create_ca_certificate_role(
            name="testrole2",
            certificate=self.TEST_CERTIFICATE,
            mount_point=self.TEST_MOUNT_POINT,
        )
        response = self.client.auth.cert.delete_certificate_role(
            name="testrole2",
            mount_point=self.TEST_MOUNT_POINT,
        )

        self.assertEqual(first=204, second=response.status_code)

    def test_configure_tls_certificate(self):
        response = self.client.auth.cert.configure_tls_certificate(
            mount_point=self.TEST_MOUNT_POINT
        )

        self.assertEqual(first=204, second=response.status_code)

    @parameterized.expand(
        [
            (TEST_ROLE_NAME, "", cert[0], cert[1], TEST_MOUNT_POINT),
            ("", "", cert[0], cert[1], TEST_MOUNT_POINT),
            ("testrole2", "", cert[0], cert[1], TEST_MOUNT_POINT),
            ("", "", "bad cert", cert[1], TEST_MOUNT_POINT),
            ("", "bad ca", cert[0], cert[1], TEST_MOUNT_POINT),
            ("", True, cert[0], cert[1], TEST_MOUNT_POINT),
            ("", False, " ", " ", TEST_MOUNT_POINT),
        ]
    )
    def test_login(self, name, cacert, cert_pem, key_pem, mount_point):
        if cacert or "bad" in [cacert, cert_pem, key_pem]:
            with self.assertRaises(exceptions.ParamValidationError):
                self.client.auth.cert.login(
                    name=name,
                    cacert=cacert,
                    cert_pem=cert_pem,
                    mount_point=mount_point,
                )
        elif (
            name != ""
            and name
            not in self.client.auth.cert.list_certificate_roles(
                mount_point=self.TEST_MOUNT_POINT,
            )["data"]["keys"]
        ):
            with self.assertRaises(exceptions.InvalidRequest):
                with self.assertRaises(OSError):
                    self.client.auth.cert.login(
                        name=name,
                        cacert=cacert,
                        cert_pem=cert_pem,
                        mount_point=mount_point,
                    )
        elif "/" not in cert_pem:
            with self.assertRaises(OSError):
                self.client.auth.cert.login(
                    name=name,
                    cacert=cacert,
                    cert_pem=cert_pem,
                    mount_point=mount_point,
                )
        else:
            response = self.client.auth.cert.login(
                name=name,
                cacert=cacert,
                cert_pem=cert_pem,
                mount_point=mount_point,
            )

            if name in [self.TEST_ROLE_NAME, ""] and (cacert, cert_pem, key_pem) == (
                "",
                self.cert[0],
                self.cert[1],
            ):
                self.assertIsInstance(response, dict)


class TestCertEnv(TestCert):
    use_env = True



from unittest import TestCase
from unittest import skipIf

from parameterized import parameterized

from hvac import exceptions
from tests.utils.hvac_integration_test_case import HvacIntegrationTestCase
from tests import utils

TEST_AUTH_PATH = "userpasswithmfa"
UNSUPPORTED_AUTH_PATH = "approle-that-can-not-have-mfa"


@skipIf(
    utils.vault_version_ge("1.11.0"),
    "Legacy MFA support dropped in Vault <1.11.0",
)
class TestLegacyMfa(HvacIntegrationTestCase, TestCase):
    mock_server_port = None

    @classmethod
    def setUpClass(cls):
        super().setUpClass()

    @classmethod
    def tearDownClass(cls):
        super().tearDownClass()

    def setUp(self):
        super().setUp()
        if "%s/" % TEST_AUTH_PATH not in self.client.sys.list_auth_methods():
            self.client.sys.enable_auth_method(
                method_type="userpass", path=TEST_AUTH_PATH
            )
        if "%s/" % UNSUPPORTED_AUTH_PATH not in self.client.sys.list_auth_methods():
            self.client.sys.enable_auth_method(
                method_type="approle", path=UNSUPPORTED_AUTH_PATH
            )

    def tearDown(self):
        super().tearDown()
        for path in [TEST_AUTH_PATH, UNSUPPORTED_AUTH_PATH]:
            self.client.sys.disable_auth_method(
                path=path,
            )

    @parameterized.expand(
        [
            ("enable mfa with supported auth method", TEST_AUTH_PATH),
            (
                "enable mfa with unsupported mfa type",
                TEST_AUTH_PATH,
                "cats",
                False,
                exceptions.ParamValidationError,
                "Unsupported mfa_type argument provided",
            ),
            (
                "enable mfa with unconfigured auth method path",
                "whats-all-this-then",
                "duo",
                False,
                exceptions.InvalidPath,
                "no handler for route",
            ),
            (
                "enable mfa with unsupported auth method type",
                UNSUPPORTED_AUTH_PATH,
                "duo",
                False,
                exceptions.InvalidPath,
                "unsupported path",
            ),
            (
                "enable mfa with unsupported auth method type forced",
                TEST_AUTH_PATH,
                "cats",
                True,
            ),
        ]
    )
    def test_configure(
        self,
        test_label,
        mount_point,
        mfa_type="duo",
        force=False,
        raises=None,
        exception_message="",
    ):
        if raises:
            with self.assertRaises(raises) as cm:
                self.client.auth.mfa.configure(
                    mount_point=mount_point,
                    mfa_type=mfa_type,
                    force=force,
                )
            self.assertIn(
                member=exception_message,
                container=str(cm.exception),
            )
        else:
            expected_status_code = 204
            configure_response = self.client.auth.mfa.configure(
                mount_point=mount_point,
                mfa_type=mfa_type,
                force=force,
            )
            self.assertEqual(
                first=expected_status_code, second=configure_response.status_code
            )

            read_config_response = self.client.auth.mfa.read_configuration(
                mount_point=mount_point,
            )
            self.assertEqual(
                first=mfa_type, second=read_config_response["data"]["type"]
            )

    @parameterized.expand(
        [
            ("read configured path", TEST_AUTH_PATH),
        ]
    )
    def test_read_configuration(self, test_label, mount_point, add_configuration=True):
        if add_configuration:
            self.client.auth.mfa.configure(
                mount_point=mount_point,
            )

        response = self.client.auth.mfa.read_configuration(
            mount_point=mount_point,
        )
        self.assertIn(
            member="data",
            container=response,
        )

    @parameterized.expand(
        [
            ("configure duo access success", TEST_AUTH_PATH),
        ]
    )
    def test_configure_duo_access(
        self,
        test_label,
        mount_point,
        host="",
        integration_key="",
        secret_key="",
        raises=None,
        exception_message="",
    ):
        if raises:
            with self.assertRaises(raises) as cm:
                self.client.auth.mfa.configure_duo_access(
                    mount_point=mount_point,
                    host=host,
                    integration_key=integration_key,
                    secret_key=secret_key,
                )
            self.assertIn(
                member=exception_message,
                container=str(cm.exception),
            )
        else:
            expected_status_code = 204
            configure_response = self.client.auth.mfa.configure_duo_access(
                mount_point=mount_point,
                host=host,
                integration_key=integration_key,
                secret_key=secret_key,
            )
            self.assertEqual(
                first=expected_status_code, second=configure_response.status_code
            )

    @parameterized.expand(
        [
            ("enable mfa with supported auth method", TEST_AUTH_PATH),
        ]
    )
    def test_configure_duo_behavior(
        self,
        test_label,
        mount_point,
        push_info="",
        user_agent="",
        username_format="%s",
        raises=None,
        exception_message="",
    ):
        if raises:
            with self.assertRaises(raises) as cm:
                self.client.auth.mfa.configure_duo_behavior(
                    mount_point=mount_point,
                    push_info=push_info,
                    user_agent=user_agent,
                    username_format=username_format,
                )
            self.assertIn(
                member=exception_message,
                container=str(cm.exception),
            )
        else:
            expected_status_code = 204
            configure_response = self.client.auth.mfa.configure_duo_behavior(
                mount_point=mount_point,
                push_info=push_info,
                user_agent=user_agent,
                username_format=username_format,
            )
            self.assertEqual(
                first=expected_status_code, second=configure_response.status_code
            )

            read_config_response = self.client.auth.mfa.read_duo_behavior_configuration(
                mount_point=mount_point,
            )
            self.assertEqual(
                first=push_info, second=read_config_response["data"]["push_info"]
            )

    @parameterized.expand(
        [
            ("read configured path", TEST_AUTH_PATH),
        ]
    )
    def test_read_duo_behavior_configuration(
        self, test_label, mount_point, add_configuration=True
    ):
        if add_configuration:
            self.client.auth.mfa.configure(
                mount_point=mount_point,
            )

        response = self.client.auth.mfa.read_duo_behavior_configuration(
            mount_point=mount_point,
        )
        self.assertIn(
            member="data",
            container=response,
        )

    @parameterized.expand(
        [
            (
                "login without duo access configured",
                False,
                exceptions.InvalidRequest,
                "Duo access credentials haven't been configured.",
            ),
        ]
    )
    def test_login_with_mfa(
        self, test_label, configure_access=True, raises=None, exception_message=""
    ):
        username = "somedude"
        password = "myverygoodpassword"

        self.client.auth.mfa.configure(
            mount_point=TEST_AUTH_PATH,
        )
        if configure_access:
            self.client.auth.mfa.configure_duo_access(
                mount_point=TEST_AUTH_PATH,
                host=f"localhost:{self.mock_server_port}",
                integration_key="an-integration-key",
                secret_key="valid-secret-key",
            )

        self.client.auth.userpass.create_or_update_user(
            username=username,
            password=password,
            policies=["defaut"],
            mount_point=TEST_AUTH_PATH,
        )
        if raises:
            with self.assertRaises(raises) as cm:
                self.client.auth.userpass.login(
                    username=username,
                    password=password,
                    mount_point=TEST_AUTH_PATH,
                )
            self.assertIn(
                member=exception_message,
                container=str(cm.exception),
            )
        else:
            # TODO: Add full userpass + MFA integration test case should a mock Duo API server become available and/or
            # if a option for toggling the following SetInsecure func on the vendor code Vault itself uses is exposed:
            # https://github.com/duosecurity/duo_api_golang/blob/master/duoapi.go#L84-L89
            pass



#!/usr/bin/env python
from ldap_test import LdapServer
from tests.utils import get_free_port


class MockLdapServer:
    ldap_url = "ldap://ldap.python-hvac.org"
    ldap_group_name = "vault-users"
    ldap_user_name = "somedude"
    ldap_user_password = "hvacrox"
    ldap_base_dc = "hvac"
    ldap_base_dn = f"dc={ldap_base_dc},dc=network"
    ldap_bind_dn = f"cn=admin,{ldap_base_dn}"
    ldap_bind_password = "notaverygoodpassword"
    ldap_users_dn = f"dc=users,{ldap_base_dn}"
    ldap_groups_ou = "groups"
    ldap_groups_dn = f"ou={ldap_groups_ou},{ldap_base_dn}"
    ldap_login_user_dn = "uid={username},{users_dn}".format(
        username=ldap_user_name, users_dn=ldap_users_dn
    )
    ldap_entries = [
        {"objectclass": "domain", "dn": ldap_users_dn, "attributes": {"dc": "users"}},
        {
            "objectclass": ["inetorgperson", "posixgroup", "top"],
            "dn": ldap_login_user_dn,
            "attributes": {"uid": ldap_user_name, "userpassword": ldap_user_password},
        },
        {
            "objectclass": "organizationalunit",
            "dn": ldap_groups_dn,
            "attributes": {
                "ou": "groups",
            },
        },
        {
            "objectclass": "groupofnames",
            "dn": "cn={cn},{groups_dn}".format(
                cn=ldap_group_name, groups_dn=ldap_groups_dn
            ),
            "attributes": {
                "cn": ldap_group_name,
                "member": ldap_login_user_dn,
            },
        },
    ]

    def __init__(self):
        self.server_port = get_free_port()
        self.ldap_server = LdapServer(
            {
                "port": self.server_port,
                "bind_dn": self.ldap_bind_dn,
                "password": self.ldap_bind_password,
                "base": {
                    "objectclass": ["domain"],
                    "dn": self.ldap_base_dn,
                    "attributes": {"dc": self.ldap_base_dc},
                },
                "entries": self.ldap_entries,
            }
        )

    @property
    def url(self):
        return f"ldap://localhost:{self.server_port}"

    def start(self):
        self.ldap_server.start()

    def stop(self):
        self.ldap_server.stop()



import logging
from unittest import TestCase

import requests_mock
from parameterized import parameterized

from hvac.adapters import JSONAdapter
from hvac.api.auth_methods import Gcp
from tests import utils


class TestGcp(TestCase):
    TEST_MOUNT_POINT = "gcp-test"

    @parameterized.expand(
        [
            (
                "success",
                dict(),
                None,
            ),
        ]
    )
    @requests_mock.Mocker()
    def test_login(self, label, test_params, raises, requests_mocker):
        role_name = "hvac"
        credentials = utils.load_config_file("example.jwt.json")
        test_policies = [
            "default",
            "dev",
            "prod",
        ]
        expected_status_code = 200
        mock_url = "http://localhost:8200/v1/auth/{mount_point}/login".format(
            mount_point=self.TEST_MOUNT_POINT,
        )
        mock_response = {
            "auth": {
                "client_token": "f33f8c72-924e-11f8-cb43-ac59d697597c",
                "accessor": "0e9e354a-520f-df04-6867-ee81cae3d42d",
                "policies": test_policies,
                "lease_duration": 2764800,
                "renewable": True,
            },
        }
        requests_mocker.register_uri(
            method="POST",
            url=mock_url,
            status_code=expected_status_code,
            json=mock_response,
        )
        gcp = Gcp(adapter=JSONAdapter())
        if raises is not None:
            with self.assertRaises(raises):
                gcp.login(
                    role=role_name,
                    jwt=credentials,
                    mount_point=self.TEST_MOUNT_POINT,
                    **test_params
                )
        else:
            login_response = gcp.login(
                role=role_name,
                jwt=credentials,
                mount_point=self.TEST_MOUNT_POINT,
                **test_params
            )
            logging.debug("login_response: %s" % login_response)
            self.assertEqual(
                first=login_response["auth"]["policies"],
                second=test_policies,
            )



import logging
from unittest import TestCase

from tests.utils.hvac_integration_test_case import HvacIntegrationTestCase


class TestWrapping(HvacIntegrationTestCase, TestCase):
    TEST_AUTH_METHOD_TYPE = "approle"
    TEST_AUTH_METHOD_PATH = "test-approle"

    def setUp(self):
        super().setUp()
        self.client.sys.enable_auth_method(
            method_type=self.TEST_AUTH_METHOD_TYPE,
            path=self.TEST_AUTH_METHOD_PATH,
        )

    def test_unwrap(self):
        self.client.write(
            path=f"auth/{self.TEST_AUTH_METHOD_PATH}/role/testrole",
        )
        result = self.client.write(
            path="auth/{path}/role/testrole/secret-id".format(
                path=self.TEST_AUTH_METHOD_PATH
            ),
            wrap_ttl="10s",
        )
        self.assertIn("token", result["wrap_info"])

        unwrap_response = self.client.sys.unwrap(result["wrap_info"]["token"])
        logging.debug("unwrap_response: %s" % unwrap_response)
        self.assertIn(member="secret_id_accessor", container=unwrap_response["data"])
        self.assertIn(member="secret_id", container=unwrap_response["data"])
